# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

__hf_hub_stable_v1_5_dict__ = lambda name: {
    "unet": {
        "config": f"https://huggingface.co/{name}/resolve/main/unet/config.json",
        "weight": f"https://huggingface.co/{name}/resolve/main/unet/diffusion_pytorch_model.bin",
    },
    "text": {
        "config": f"https://huggingface.co/{name}/resolve/main/text_encoder/config.json",
        "vocab": f"https://huggingface.co/{name}/resolve/main/tokenizer/vocab.json",
        "merge": f"https://huggingface.co/{name}/resolve/main/tokenizer/merges.txt",
        "weight": f"https://huggingface.co/{name}/resolve/main/text_encoder/pytorch_model.bin",
    },
    "vae": {
        "config": f"https://huggingface.co/{name}/resolve/main/vae/config.json",
        "weight": f"https://huggingface.co/{name}/resolve/main/vae/diffusion_pytorch_model.bin",
    },
    "scheduler": f"https://huggingface.co/{name}/resolve/main/scheduler/scheduler_config.json",
}

__hf_hub_stable_v1_5_safetensors_dict__ = lambda name: {
    "unet": {
        "config": f"https://huggingface.co/{name}/resolve/main/unet/config.json",
        "weight": f"https://huggingface.co/{name}/resolve/main/unet/diffusion_pytorch_model.safetensors",
    },
    "text": {
        "config": f"https://huggingface.co/{name}/resolve/main/text_encoder/config.json",
        "vocab": f"https://huggingface.co/{name}/resolve/main/tokenizer/vocab.json",
        "merge": f"https://huggingface.co/{name}/resolve/main/tokenizer/merges.txt",
        "weight": f"https://huggingface.co/{name}/resolve/main/text_encoder/model.safetensors",
    },
    "vae": {
        "config": f"https://huggingface.co/{name}/resolve/main/vae/config.json",
        "weight": f"https://huggingface.co/{name}/resolve/main/vae/diffusion_pytorch_model.safetensors",
    },
    "scheduler": f"https://huggingface.co/{name}/resolve/main/scheduler/scheduler_config.json",
}

__hf_hub_stable_v2_dict__ = __hf_hub_stable_v1_5_dict__
__hf_hub_stable_v2_safetensors_dict__ = __hf_hub_stable_v1_5_safetensors_dict__
__hf_hub_stable_v2_1_dict__ = __hf_hub_stable_v1_5_dict__
__hf_hub_stable_v2_1_safetensors_dict__ = __hf_hub_stable_v1_5_safetensors_dict__

__hf_hub_stable_xl_dict__ = lambda name: {
    "unet": {
        "config": f"https://huggingface.co/{name}/resolve/main/unet/config.json",
        "weight": f"https://huggingface.co/{name}/resolve/main/unet/diffusion_pytorch_model.bin",
    },
    "text": {
        "config": f"https://huggingface.co/{name}/resolve/main/text_encoder/config.json",
        "vocab": f"https://huggingface.co/{name}/resolve/main/tokenizer/vocab.json",
        "merge": f"https://huggingface.co/{name}/resolve/main/tokenizer/merges.txt",
        "weight": f"https://huggingface.co/{name}/resolve/main/text_encoder/pytorch_model.bin",
    },
    "text2": {
        "config": f"https://huggingface.co/{name}/resolve/main/text_encoder_2/config.json",
        "vocab": f"https://huggingface.co/{name}/resolve/main/tokenizer_2/vocab.json",
        "merge": f"https://huggingface.co/{name}/resolve/main/tokenizer_2/merges.txt",
        "weight": f"https://huggingface.co/{name}/resolve/main/text_encoder_2/pytorch_model.bin",
    },
    "vae": {
        "config": f"https://huggingface.co/{name}/resolve/main/vae/config.json",
        "weight": f"https://huggingface.co/{name}/resolve/main/vae/diffusion_pytorch_model.bin",
    },
    "scheduler": f"https://huggingface.co/{name}/resolve/main/scheduler/scheduler_config.json",
}

__hf_hub_stable_xl_safetensors_dict__ = lambda name: {
    "unet": {
        "config": f"https://huggingface.co/{name}/resolve/main/unet/config.json",
        "weight": f"https://huggingface.co/{name}/resolve/main/unet/diffusion_pytorch_model.fp16.safetensors",
    },
    "text": {
        "config": f"https://huggingface.co/{name}/resolve/main/text_encoder/config.json",
        "vocab": f"https://huggingface.co/{name}/resolve/main/tokenizer/vocab.json",
        "merge": f"https://huggingface.co/{name}/resolve/main/tokenizer/merges.txt",
        "weight": f"https://huggingface.co/{name}/resolve/main/text_encoder/model.fp16.safetensors",
    },
    "text2": {
        "config": f"https://huggingface.co/{name}/resolve/main/text_encoder_2/config.json",
        "vocab": f"https://huggingface.co/{name}/resolve/main/tokenizer_2/vocab.json",
        "merge": f"https://huggingface.co/{name}/resolve/main/tokenizer_2/merges.txt",
        "weight": f"https://huggingface.co/{name}/resolve/main/text_encoder_2/model.fp16.safetensors",
    },
    "vae": {
        "config": f"https://huggingface.co/{name}/resolve/main/vae/config.json",
        "weight": f"https://huggingface.co/{name}/resolve/main/vae/diffusion_pytorch_model.fp16.safetensors",
    },
    "scheduler": f"https://huggingface.co/{name}/resolve/main/scheduler/scheduler_config.json",
}

__hf_hub_stable_3_safetensors_dict__ = lambda name: {
    "transformer": {
        "config": f"https://huggingface.co/{name}/resolve/main/transformer/config.json",
        "weight": f"https://huggingface.co/{name}/resolve/main/transformer/diffusion_pytorch_model.safetensors",
    },
    "text": {
        "config": f"https://huggingface.co/{name}/resolve/main/text_encoder/config.json",
        "vocab": f"https://huggingface.co/{name}/resolve/main/tokenizer/vocab.json",
        "merge": f"https://huggingface.co/{name}/resolve/main/tokenizer/merges.txt",
        "weight": f"https://huggingface.co/{name}/resolve/main/text_encoder/model.safetensors",
    },
    "text2": {
        "config": f"https://huggingface.co/{name}/resolve/main/text_encoder_2/config.json",
        "vocab": f"https://huggingface.co/{name}/resolve/main/tokenizer_2/vocab.json",
        "merge": f"https://huggingface.co/{name}/resolve/main/tokenizer_2/merges.txt",
        "weight": f"https://huggingface.co/{name}/resolve/main/text_encoder_2/model.safetensors",
    },
    "text3": {
        "config": f"https://huggingface.co/{name}/resolve/main/text_encoder_3/config.json",
        "vocab": f"https://huggingface.co/{name}/resolve/main/tokenizer_3/spiece.model",
        "weight": [
            f"https://huggingface.co/{name}/resolve/main/text_encoder_3/model-{str(i).rjust(5, '0')}-of-00002.safetensors"
            for i in range(1, 3)
        ],
    },
    "vae": {
        "config": f"https://huggingface.co/{name}/resolve/main/vae/config.json",
        "weight": f"https://huggingface.co/{name}/resolve/main/vae/diffusion_pytorch_model.safetensors",
    },
    "scheduler": f"https://huggingface.co/{name}/resolve/main/scheduler/scheduler_config.json",
}

__hf_hub_controlnet_dict__ = lambda name: {
    "controlnet": {
        "config": f"https://huggingface.co/{name}/resolve/main/config.json",
        "weight": f"https://huggingface.co/{name}/resolve/main/diffusion_pytorch_model.bin",
    }
}

__hf_hub_controlnet_safetensors_dict__ = lambda name: {
    "controlnet": {
        "config": f"https://huggingface.co/{name}/resolve/main/config.json",
        "weight": f"https://huggingface.co/{name}/resolve/main/diffusion_pytorch_model.safetensors",
    }
}

__hf_hub_adapter_safetensors_dict__ = lambda name: {
    "adapter": {
        "config": f"https://huggingface.co/{name}/resolve/main/config.json",
        "weight": f"https://huggingface.co/{name}/resolve/main/diffusion_pytorch_model.safetensors",
    }
}

__hf_hub_stable_video_safetensors_dict__ = lambda name: {
    "unet": {
        "config": f"https://huggingface.co/{name}/resolve/main/unet/config.json",
        "weight": f"https://huggingface.co/{name}/resolve/main/unet/diffusion_pytorch_model.safetensors",
    },
    "image": {
        "config": f"https://huggingface.co/{name}/resolve/main/image_encoder/config.json",
        "vision_config": f"https://huggingface.co/{name}/resolve/main/feature_extractor/preprocessor_config.json",
        "weight": f"https://huggingface.co/{name}/resolve/main/image_encoder/model.safetensors",
    },
    "vae": {
        "config": f"https://huggingface.co/{name}/resolve/main/vae/config.json",
        "weight": f"https://huggingface.co/{name}/resolve/main/vae/diffusion_pytorch_model.safetensors",
    },
    "scheduler": f"https://huggingface.co/{name}/resolve/main/scheduler/scheduler_config.json",
}

pretrained_stable_infos = {
    "stable-v1.5": __hf_hub_stable_v1_5_dict__("runwayml/stable-diffusion-v1-5"),
    "stable-v1.5-realistic-v5.1-no-vae": __hf_hub_stable_v1_5_safetensors_dict__(
        "SG161222/Realistic_Vision_V5.1_noVAE"
    ),
    "stable-v1.5-realistic-v5.1": __hf_hub_stable_v1_5_dict__(
        "stablediffusionapi/realistic-vision-v51"
    ),
    "stable-v1.5-cyber-realistic": __hf_hub_stable_v1_5_safetensors_dict__(
        "Yntec/CyberRealistic"
    ),
    "stable-v1.5-majicmix-realistic-v6": __hf_hub_stable_v1_5_dict__(
        "digiplay/majicMIX_realistic_v6"
    ),
    "stable-v1.5-nitrosocke-ghibli": __hf_hub_stable_v1_5_dict__(
        "nitrosocke/Ghibli-Diffusion"
    ),
    "stable-v1.5-inpainting": __hf_hub_stable_v1_5_dict__(
        "runwayml/stable-diffusion-inpainting"
    ),
    "stable-v1.5-x4-upscaler": __hf_hub_stable_v1_5_dict__(
        "stabilityai/stable-diffusion-x4-upscaler"
    ),
    "stable-v2": __hf_hub_stable_v2_dict__("stabilityai/stable-diffusion-2"),
    "stable-v2.1": __hf_hub_stable_v2_1_dict__("stabilityai/stable-diffusion-2-1"),
    "stable-xl-base": __hf_hub_stable_xl_safetensors_dict__(
        "stabilityai/stable-diffusion-xl-base-1.0"
    ),
    "stable-xl-base-vae-fp16": {
        **__hf_hub_stable_xl_safetensors_dict__(
            "stabilityai/stable-diffusion-xl-base-1.0"
        ),
        **{
            "vae": {
                "config": "https://huggingface.co/madebyollin/sdxl-vae-fp16-fix/resolve/main/config.json",
                "weight": "https://huggingface.co/madebyollin/sdxl-vae-fp16-fix/resolve/main/diffusion_pytorch_model.safetensors",
            },
        },
    },
    "stable-xl-turbo": __hf_hub_stable_xl_safetensors_dict__("stabilityai/sdxl-turbo"),
    "stable-xl-opendalle-v1.1": __hf_hub_stable_xl_safetensors_dict__(
        "dataautogpt3/OpenDalleV1.1"
    ),
    "stable-xl-realvis-v3.0": __hf_hub_stable_xl_safetensors_dict__(
        "SG161222/RealVisXL_V3.0"
    ),
    "stable-xl-juggernaut-v8": __hf_hub_stable_xl_dict__(
        "RunDiffusion/Juggernaut-XL-v8"
    ),
    "stable-xl-playground-v2-aesthetic": __hf_hub_stable_xl_safetensors_dict__(
        "playgroundai/playground-v2-1024px-aesthetic"
    ),
    "stable-v3-medium": __hf_hub_stable_3_safetensors_dict__(
        "ckpt/stable-diffusion-3-medium-diffusers"
    ),
    "stable-video-diffusion-img2vid-xt": __hf_hub_stable_video_safetensors_dict__(
        "stabilityai/stable-video-diffusion-img2vid-xt"
    ),
    "stable-video-diffusion-img2vid-xt-1-1": __hf_hub_stable_video_safetensors_dict__(
        "vdo/stable-video-diffusion-img2vid-xt-1-1"
    ),
}

pretrained_stable_extensions_infos = {
    "stable-v1.5-controlnet-canny": __hf_hub_controlnet_dict__(
        "lllyasviel/sd-controlnet-canny"
    ),
    "stable-v1.5-controlnet-openpose": __hf_hub_controlnet_dict__(
        "lllyasviel/sd-controlnet-openpose"
    ),
    "stable-v1.5-controlnet-softedge": __hf_hub_controlnet_dict__(
        "lllyasviel/control_v11p_sd15_softedge",
    ),
    "stable-v1.5-controlnet-inpainting": __hf_hub_controlnet_dict__(
        "lllyasviel/control_v11p_sd15_inpaint"
    ),
    "stable-v1.5-controlnet-depth": __hf_hub_controlnet_dict__(
        "lllyasviel/control_v11f1p_sd15_depth"
    ),
    "stable-xl-controlnet-canny": __hf_hub_controlnet_dict__(
        "diffusers/controlnet-canny-sdxl-1.0"
    ),
    "stable-xl-controlnet-softedge-dexined": __hf_hub_controlnet_dict__(
        "SargeZT/controlnet-sd-xl-1.0-softedge-dexined"
    ),
    "stable-xl-controlnet-depth": __hf_hub_controlnet_dict__(
        "diffusers/controlnet-depth-sdxl-1.0"
    ),
    "stable-xl-controlnet-softedge": __hf_hub_controlnet_dict__(
        "diffusers/controlnet-softedge-sdxl-1.0"
    ),
    "stable-xl-controlnet-depth-small": __hf_hub_controlnet_dict__(
        "diffusers/controlnet-depth-sdxl-1.0-small"
    ),
    "stable-xl-adapter-canny": __hf_hub_adapter_safetensors_dict__(
        "TencentARC/t2i-adapter-canny-sdxl-1.0"
    ),
    "stable-xl-adapter-sketch": __hf_hub_adapter_safetensors_dict__(
        "TencentARC/t2i-adapter-sketch-sdxl-1.0"
    ),
    "stable-xl-adapter-depth": __hf_hub_adapter_safetensors_dict__(
        "TencentARC/t2i-adapter-depth-zoe-sdxl-1.0"
    ),
    "stable-xl-adapter-depth-midas": __hf_hub_adapter_safetensors_dict__(
        "TencentARC/t2i-adapter-depth-midas-sdxl-1.0"
    ),
    "stable-xl-adapter-openpose": __hf_hub_adapter_safetensors_dict__(
        "TencentARC/t2i-adapter-openpose-sdxl-1.0"
    ),
    "stable-xl-refiner-1.0": {
        "refiner": {
            "unet": {
                "config": f"https://huggingface.co/stabilityai/stable-diffusion-xl-refiner-1.0/resolve/main/unet/config.json",
                "weight": f"https://huggingface.co/stabilityai/stable-diffusion-xl-refiner-1.0/resolve/main/unet/diffusion_pytorch_model.fp16.safetensors",
            },
            "text2": {
                "config": f"https://huggingface.co/stabilityai/stable-diffusion-xl-refiner-1.0/resolve/main/text_encoder_2/config.json",
                "vocab": f"https://huggingface.co/stabilityai/stable-diffusion-xl-refiner-1.0/resolve/main/tokenizer_2/vocab.json",
                "merge": f"https://huggingface.co/stabilityai/stable-diffusion-xl-refiner-1.0/resolve/main/tokenizer_2/merges.txt",
                "weight": f"https://huggingface.co/stabilityai/stable-diffusion-xl-refiner-1.0/resolve/main/text_encoder_2/model.fp16.safetensors",
            },
            "vae": {
                "config": f"https://huggingface.co/stabilityai/stable-diffusion-xl-refiner-1.0/resolve/main/vae/config.json",
                "weight": f"https://huggingface.co/stabilityai/stable-diffusion-xl-refiner-1.0/resolve/main/vae/diffusion_pytorch_model.fp16.safetensors",
            },
            "scheduler": f"https://huggingface.co/stabilityai/stable-diffusion-xl-refiner-1.0/resolve/main/scheduler/scheduler_config.json",
        },
    },
    "stable-xl-vae-fp16": {
        "vae": {
            "config": "https://huggingface.co/madebyollin/sdxl-vae-fp16-fix/resolve/main/config.json",
            "weight": "https://huggingface.co/madebyollin/sdxl-vae-fp16-fix/resolve/main/diffusion_pytorch_model.safetensors",
        },
    },
    "stable-v3-controlnet-canny": __hf_hub_controlnet_safetensors_dict__(
        "InstantX/SD3-Controlnet-Canny"
    ),
}

from unitorch.cli.models.diffusion_utils import load_weight

from unitorch.cli.models.diffusers.modeling_stable import (
    StableForText2ImageGeneration,
    StableForImage2ImageGeneration,
    StableForImageInpainting,
    StableForImageResolution,
)
from unitorch.cli.models.diffusers.modeling_stable_xl import (
    StableXLForText2ImageGeneration,
    StableXLForImage2ImageGeneration,
    StableXLForImageInpainting,
)
from unitorch.cli.models.diffusers.modeling_stable_3 import (
    Stable3ForText2ImageGeneration,
    Stable3ForImage2ImageGeneration,
)
from unitorch.cli.models.diffusers.modeling_controlnet import (
    ControlNetForText2ImageGeneration,
    ControlNetForImage2ImageGeneration,
    ControlNetForImageInpainting,
)
from unitorch.cli.models.diffusers.modeling_controlnet_xl import (
    ControlNetXLForText2ImageGeneration,
    ControlNetXLForImage2ImageGeneration,
    ControlNetXLForImageInpainting,
)
from unitorch.cli.models.diffusers.modeling_controlnet_3 import (
    ControlNet3ForText2ImageGeneration,
)
from unitorch.cli.models.diffusers.modeling_adapter_xl import (
    StableXLAdapterForText2ImageGeneration,
)
from unitorch.cli.models.diffusers.processing_stable import StableProcessor
from unitorch.cli.models.diffusers.processing_stable_xl import StableXLProcessor
from unitorch.cli.models.diffusers.processing_controlnet import ControlNetProcessor
from unitorch.cli.models.diffusers.processing_controlnet_xl import ControlNetXLProcessor
from unitorch.cli.models.diffusers.processing_controlnet_3 import ControlNet3Processor
from unitorch.cli.models.diffusers.processing_adapter_xl import AdapterXLProcessor



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import torch
from PIL import Image
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from transformers import PreTrainedTokenizer
from transformers.image_utils import to_numpy_array, ChannelDimension
from transformers.image_transforms import to_channel_dimension_format
from transformers.image_processing_utils import BaseImageProcessor
from unitorch.models import GenericOutputs
from unitorch.utils import pop_value, truncate_sequence_pair


class HfTextGenerationProcessor:
    """
    Processor for text generation tasks.

    Args:
        tokenizer (PreTrainedTokenizer): The tokenizer to use for text generation.
        max_seq_length (int, optional): Maximum sequence length. Defaults to 128.
        max_gen_seq_length (int, optional): Maximum generated sequence length. Defaults to 48.
    """

    def __init__(
        self,
        tokenizer: PreTrainedTokenizer,
        max_seq_length: Optional[int] = 128,
        max_gen_seq_length: Optional[int] = 48,
    ):
        self.tokenizer = tokenizer
        self.max_seq_length = max_seq_length
        self.max_gen_seq_length = max_gen_seq_length
        self.pad_token = self.tokenizer.pad_token
        self.bos_token = self.tokenizer.bos_token
        self.eos_token = self.tokenizer.eos_token
        self.mask_token = self.tokenizer.mask_token
        self.pad_token_id = self.tokenizer.pad_token_id
        self.vocab_size = len(self.tokenizer.get_vocab())

    def generation_inputs(
        self,
        text: str,
        max_seq_length: Optional[int] = None,
    ):
        """
        Generate inputs for text generation.

        Args:
            text (str): The input text.
            max_seq_length (int, optional): Maximum sequence length. Defaults to None.

        Returns:
            GenericOutputs: The generated input tokens and attention mask.
        """
        max_seq_length = pop_value(max_seq_length, self.max_seq_length)
        tokens = self.tokenizer.tokenize(str(text))
        tokens = tokens[: max_seq_length - 2]
        tokens = [self.bos_token] + tokens + [self.eos_token]
        input_ids = self.tokenizer.convert_tokens_to_ids(tokens)
        input_ids = input_ids[:max_seq_length]
        attention_mask = [1] * len(input_ids)

        padding = [0] * (max_seq_length - len(input_ids))
        input_ids += [self.pad_token_id] * len(padding)
        attention_mask += padding

        assert len(input_ids) == max_seq_length
        assert len(attention_mask) == max_seq_length

        return GenericOutputs(
            input_ids=torch.tensor(input_ids, dtype=torch.long),
            attention_mask=torch.tensor(attention_mask, dtype=torch.long),
        )

    def generation_labels(
        self,
        text: str,
        max_gen_seq_length: Optional[int] = None,
    ):
        """
        Generate labels for text generation.

        Args:
            text (str): The input text.
            max_gen_seq_length (int, optional): Maximum generated sequence length. Defaults to None.

        Returns:
            GenericOutputs: The generated label tokens and attention mask.
        """
        max_gen_seq_length = pop_value(max_gen_seq_length, self.max_gen_seq_length)
        tokens = self.tokenizer.tokenize(str(text))
        tokens = tokens[: max_gen_seq_length - 2]
        tokens = [self.bos_token] + tokens + [self.eos_token]
        input_ids = self.tokenizer.convert_tokens_to_ids(tokens)
        input_ids = input_ids[1:max_gen_seq_length]
        attention_mask = [1] * len(input_ids)

        padding = [0] * (max_gen_seq_length - len(input_ids))
        input_ids += [self.pad_token_id] * len(padding)
        attention_mask += padding

        assert len(input_ids) == max_gen_seq_length

        return GenericOutputs(
            input_ids=torch.tensor(input_ids, dtype=torch.long),
            attention_mask=torch.tensor(attention_mask, dtype=torch.long),
        )

    def generation(
        self,
        text: str,
        text_pair: str,
        max_seq_length: Optional[int] = None,
        max_gen_seq_length: Optional[int] = None,
    ):
        """
        Generate inputs, labels, and tokens for text generation.

        Args:
            text (str): The input text.
            text_pair (str): The paired text.
            max_seq_length (int, optional): Maximum sequence length. Defaults to None.
            max_gen_seq_length (int, optional): Maximum generated sequence length. Defaults to None.

        Returns:
            GenericOutputs: The generated input tokens, attention masks, label tokens, and attention masks.
        """
        max_seq_length = pop_value(max_seq_length, self.max_seq_length)
        max_gen_seq_length = pop_value(max_gen_seq_length, self.max_gen_seq_length)

        tokens = self.generation_inputs(text, max_seq_length)
        tokens_pair = self.generation_inputs(text_pair, max_gen_seq_length)
        labels = self.generation_labels(text_pair, max_gen_seq_length)

        return GenericOutputs(
            input_ids=tokens.input_ids,
            attention_mask=tokens.attention_mask,
            input_ids_pair=tokens_pair.input_ids,
            attention_mask_pair=tokens_pair.attention_mask,
            input_ids_label=labels.input_ids,
            attention_mask_label=labels.attention_mask,
        )

    def detokenize(
        self,
        sequences: torch.Tensor,
        skip_special_tokens: Optional[bool] = True,
    ):
        """
        Detokenize the sequences.

        Args:
            sequences (torch.Tensor): The sequences to detokenize.
            skip_special_tokens (bool, optional): Whether to skip special tokens. Defaults to True.

        Returns:
            list: The detokenized sequences.
        """
        if sequences.dim() == 3:
            _, num_return_sequences, sequences_length = sequences.size()
            sequences = sequences.reshape(-1, sequences_length).clamp_max(
                self.vocab_size
            )
            sequences = sequences.clamp_min(0)
            sequences[sequences == self.vocab_size] = self.pad_token_id
            decode_tokens = self.tokenizer.batch_decode(
                sequences,
                skip_special_tokens=skip_special_tokens,
            )
            decode_tokens = [
                decode_tokens[i : i + num_return_sequences]
                for i in range(0, len(decode_tokens), num_return_sequences)
            ]
        elif sequences.dim() == 2:
            sequences = sequences.clamp_min(0).clamp_max(self.vocab_size)
            sequences[sequences == self.vocab_size] = self.pad_token_id
            decode_tokens = self.tokenizer.batch_decode(
                sequences,
                skip_special_tokens=skip_special_tokens,
            )
        else:
            raise ValueError(f"Can't decode the tensor with shape {sequences.shape}")

        return decode_tokens


class HfTextClassificationProcessor:
    """
    Processor for text classification tasks.

    Args:
        tokenizer (PreTrainedTokenizer): The tokenizer to use for text classification.
        max_seq_length (int, optional): Maximum sequence length. Defaults to 128.
        source_type_id (int, optional): Source type ID. Defaults to 0.
        target_type_id (int, optional): Target type ID. Defaults to 1.
        position_start_id (int, optional): Start position ID. Defaults to 0.
    """

    def __init__(
        self,
        tokenizer: PreTrainedTokenizer,
        max_seq_length: Optional[int] = 128,
        source_type_id: Optional[int] = 0,
        target_type_id: Optional[int] = 1,
        position_start_id: Optional[int] = 0,
    ):
        self.tokenizer = tokenizer
        self.max_seq_length = max_seq_length
        self.pad_token = self.tokenizer.pad_token
        self.sep_token = self.tokenizer.sep_token
        self.cls_token = self.tokenizer.cls_token
        self.mask_token = self.tokenizer.mask_token
        self.pad_token_id = self.tokenizer.pad_token_id
        self.source_type_id = source_type_id
        self.target_type_id = target_type_id
        self.position_start_id = position_start_id

    def classification(
        self,
        text: str,
        text_pair: Optional[str] = None,
        max_seq_length: Optional[int] = None,
    ):
        """
        Generate inputs for text classification.

        Args:
            text (str): The input text.
            text_pair (str, optional): The paired text. Defaults to None.
            max_seq_length (int, optional): Maximum sequence length. Defaults to None.

        Returns:
            GenericOutputs: The generated input tokens, token type IDs, attention mask, and position IDs.
        """
        max_seq_length = pop_value(
            max_seq_length,
            self.max_seq_length,
        )

        tokens = self.tokenizer.tokenize(str(text))
        if text_pair is None:
            tokens = tokens[: max_seq_length - 2]
            tokens = [self.cls_token] + tokens + [self.sep_token]
            input_ids = self.tokenizer.convert_tokens_to_ids(tokens)
            token_type_ids = [self.source_type_id] * len(input_ids)
            attention_mask = [1] * len(input_ids)
        else:
            tokens_pair = self.tokenizer.tokenize(str(text_pair))
            truncate_sequence_pair(tokens, tokens_pair, max_seq_length - 3)
            token_type_ids = (
                [self.source_type_id]
                + [self.source_type_id] * len(tokens)
                + [self.source_type_id]
                + [self.target_type_id] * len(tokens_pair)
                + [self.target_type_id]
            )
            tokens = (
                [self.cls_token]
                + tokens
                + [self.sep_token]
                + tokens_pair
                + [self.sep_token]
            )
            input_ids = self.tokenizer.convert_tokens_to_ids(tokens)
            attention_mask = [1] * len(input_ids)

        padding = [0] * (max_seq_length - len(input_ids))
        input_ids += len(padding) * [self.pad_token_id]
        attention_mask += padding
        token_type_ids += len(padding) * [self.target_type_id]

        assert len(input_ids) == max_seq_length
        assert len(attention_mask) == max_seq_length
        assert len(token_type_ids) == max_seq_length
        return GenericOutputs(
            input_ids=torch.tensor(input_ids, dtype=torch.long),
            token_type_ids=torch.tensor(token_type_ids, dtype=torch.long),
            attention_mask=torch.tensor(attention_mask, dtype=torch.long),
            position_ids=torch.tensor(
                list(
                    range(
                        self.position_start_id,
                        self.position_start_id + max_seq_length,
                    )
                ),
                dtype=torch.long,
            ),
        )


class HfImageClassificationProcessor:
    """
    Processor for image classification tasks.
    """

    def __init__(
        self,
        vision_processor: BaseImageProcessor,
    ):
        """
        Initialize the HfImageClassificationProcessor.

        Args:
            vision_processor (BaseImageProcessor): The vision processor object used for image transformations.
        """
        self.vision_processor = vision_processor

        self.size = getattr(self.vision_processor, "size", None)

        self.resample = getattr(self.vision_processor, "resample", None)

        self.crop_size = getattr(self.vision_processor, "crop_size", None)
        self.pad_size = getattr(self.vision_processor, "pad_size", None)

        self.rescale_factor = getattr(self.vision_processor, "rescale_factor", None)

        self.image_mean = getattr(self.vision_processor, "image_mean", None)
        self.image_std = getattr(self.vision_processor, "image_std", None)

    def classification(
        self,
        image: Union[Image.Image, str],
    ):
        """
        Perform image classification on the given image.

        Args:
            image (Image.Image): The input image.

        Returns:
            GenericOutputs: The output of the image classification, including pixel values.
        """
        if isinstance(image, str):
            image = Image.open(image)

        if self.size is not None:
            image = self.vision_processor.resize(
                image=to_numpy_array(image.convert("RGB")),
                size=self.size,
                resample=self.resample,
            )

        if self.crop_size is not None:
            image = self.vision_processor.center_crop(
                image,
                size=self.crop_size,
            )

        if self.rescale_factor is not None:
            image = self.vision_processor.rescale(
                image,
                self.rescale_factor,
            )

        if self.image_mean is not None and self.image_std is not None:
            image = self.vision_processor.normalize(
                image=image,
                mean=self.image_mean,
                std=self.image_std,
            )

        if self.pad_size is not None:
            image = self.vision_processor.pad_image(
                image,
                size=self.pad_size,
            )

        image = to_channel_dimension_format(image, ChannelDimension.FIRST)

        return GenericOutputs(
            pixel_values=torch.tensor(image),
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import io
import torch
import gc
import gradio as gr
from PIL import Image
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from unitorch.cli import CoreConfigureParser, GenericWebUI
from unitorch.cli import register_webui
from unitorch.cli.models.diffusers import (
    pretrained_stable_infos,
    pretrained_stable_extensions_infos,
)
from unitorch.cli.pipelines.stable_xl import StableXLForText2ImageGenerationPipeline
from unitorch.cli.pipelines.stable_xl import controlnet_processes, adapter_processes
from unitorch.cli.webuis import (
    supported_scheduler_names,
    matched_pretrained_names,
    create_element,
    create_accordion,
    create_row,
    create_column,
    create_group,
    create_tab,
    create_tabs,
    create_blocks,
    create_pretrain_layout,
    create_controlnet_layout,
    create_adapter_layout,
    create_lora_layout,
    create_freeu_layout,
)
from unitorch.cli.webuis import SimpleWebUI


class StableXLText2ImageWebUI(SimpleWebUI):
    pretrained_names = list(pretrained_stable_infos.keys())
    supported_pretrained_names = matched_pretrained_names(
        pretrained_names, "^stable-xl-"
    )
    pretrained_extension_names = list(pretrained_stable_extensions_infos.keys())
    supported_controlnet_names = matched_pretrained_names(
        pretrained_extension_names, "^stable-xl-controlnet-"
    )
    supported_controlnet_process_names = list(controlnet_processes.keys())
    supported_adapter_names = matched_pretrained_names(
        pretrained_extension_names, "^stable-xl-adapter-"
    )
    supported_adapter_process_names = list(adapter_processes.keys())
    supported_lora_names = matched_pretrained_names(
        pretrained_extension_names, "^stable-xl-lora-"
    )
    supported_schedulers = supported_scheduler_names

    def __init__(self, config: CoreConfigureParser):
        self._pipe = None if not hasattr(self, "_pipe") else self._pipe
        self._status = "Stopped" if self._pipe is None else "Running"
        if len(self.supported_pretrained_names) == 0:
            raise ValueError("No supported pretrained models found.")
        self._name = self.supported_pretrained_names[0]

        # create elements
        pretrain_layout_group = create_pretrain_layout(
            self.supported_pretrained_names, self._name
        )
        name, status, start, stop, pretrain_layout = (
            pretrain_layout_group.name,
            pretrain_layout_group.status,
            pretrain_layout_group.start,
            pretrain_layout_group.stop,
            pretrain_layout_group.layout,
        )

        prompt = create_element(
            "text", "Input Prompt", lines=3, placeholder="Prompt", show_label=False
        )
        negative_prompt = create_element(
            "text",
            "Input Negative Prompt",
            lines=3,
            placeholder="Negative Prompt",
            show_label=False,
        )
        scheduler = create_element(
            "dropdown",
            "Sampling Method",
            values=self.supported_schedulers,
            default=self.supported_schedulers[0],
        )
        steps = create_element(
            "slider", "Diffusion Steps", min_value=1, max_value=100, step=1, default=25
        )
        height = create_element(
            "slider", "Image Height", min_value=1, max_value=2048, step=1, default=1024
        )
        width = create_element(
            "slider", "Image Width", min_value=1, max_value=2048, step=1, default=1024
        )

        guidance_scale = create_element(
            "slider", "Guidance Scale", min_value=0, max_value=10, step=0.1, default=7.5
        )

        seed = create_element(
            "slider", "Seed", min_value=0, max_value=9999, step=1, default=42
        )

        freeu_layout_group = create_freeu_layout()
        s1, s2, b1, b2, freeu_layout = (
            freeu_layout_group.s1,
            freeu_layout_group.s2,
            freeu_layout_group.b1,
            freeu_layout_group.b2,
            freeu_layout_group.layout,
        )

        ## extensions
        controlnet_layout_group = create_controlnet_layout(
            self.supported_controlnet_names,
            self.supported_controlnet_process_names,
            num_controlnets=5,
        )
        controlnets = controlnet_layout_group.controlnets
        controlnet_layout = controlnet_layout_group.layout
        controlnet0, controlnet1, controlnet2, controlnet3, controlnet4 = controlnets
        (
            controlnet0_input_image,
            controlnet0_output_image,
            controlnet0_checkpoint,
            controlnet0_guidance_scale,
            controlnet0_process,
        ) = (
            controlnet0.input_image,
            controlnet0.output_image,
            controlnet0.checkpoint,
            controlnet0.guidance_scale,
            controlnet0.process,
        )
        (
            controlnet1_input_image,
            controlnet1_output_image,
            controlnet1_checkpoint,
            controlnet1_guidance_scale,
            controlnet1_process,
        ) = (
            controlnet1.input_image,
            controlnet1.output_image,
            controlnet1.checkpoint,
            controlnet1.guidance_scale,
            controlnet1.process,
        )
        (
            controlnet2_input_image,
            controlnet2_output_image,
            controlnet2_checkpoint,
            controlnet2_guidance_scale,
            controlnet2_process,
        ) = (
            controlnet2.input_image,
            controlnet2.output_image,
            controlnet2.checkpoint,
            controlnet2.guidance_scale,
            controlnet2.process,
        )
        (
            controlnet3_input_image,
            controlnet3_output_image,
            controlnet3_checkpoint,
            controlnet3_guidance_scale,
            controlnet3_process,
        ) = (
            controlnet3.input_image,
            controlnet3.output_image,
            controlnet3.checkpoint,
            controlnet3.guidance_scale,
            controlnet3.process,
        )
        (
            controlnet4_input_image,
            controlnet4_output_image,
            controlnet4_checkpoint,
            controlnet4_guidance_scale,
            controlnet4_process,
        ) = (
            controlnet4.input_image,
            controlnet4.output_image,
            controlnet4.checkpoint,
            controlnet4.guidance_scale,
            controlnet4.process,
        )
        adapter_layout_group = create_adapter_layout(
            self.supported_adapter_names,
            self.supported_adapter_process_names,
            num_adapters=5,
        )
        adapters = adapter_layout_group.adapters
        adapter_layout = adapter_layout_group.layout
        adapter0, adapter1, adapter2, adapter3, adapter4 = adapters
        (
            adapter0_input_image,
            adapter0_output_image,
            adapter0_checkpoint,
            adapter0_guidance_scale,
            adapter0_process,
        ) = (
            adapter0.input_image,
            adapter0.output_image,
            adapter0.checkpoint,
            adapter0.guidance_scale,
            adapter0.process,
        )
        (
            adapter1_input_image,
            adapter1_output_image,
            adapter1_checkpoint,
            adapter1_guidance_scale,
            adapter1_process,
        ) = (
            adapter1.input_image,
            adapter1.output_image,
            adapter1.checkpoint,
            adapter1.guidance_scale,
            adapter1.process,
        )
        (
            adapter2_input_image,
            adapter2_output_image,
            adapter2_checkpoint,
            adapter2_guidance_scale,
            adapter2_process,
        ) = (
            adapter2.input_image,
            adapter2.output_image,
            adapter2.checkpoint,
            adapter2.guidance_scale,
            adapter2.process,
        )
        (
            adapter3_input_image,
            adapter3_output_image,
            adapter3_checkpoint,
            adapter3_guidance_scale,
            adapter3_process,
        ) = (
            adapter3.input_image,
            adapter3.output_image,
            adapter3.checkpoint,
            adapter3.guidance_scale,
            adapter3.process,
        )
        (
            adapter4_input_image,
            adapter4_output_image,
            adapter4_checkpoint,
            adapter4_guidance_scale,
            adapter4_process,
        ) = (
            adapter4.input_image,
            adapter4.output_image,
            adapter4.checkpoint,
            adapter4.guidance_scale,
            adapter4.process,
        )
        lora_layout_group = create_lora_layout(self.supported_lora_names, num_loras=5)
        loras = lora_layout_group.loras
        lora_layout = lora_layout_group.layout
        lora0, lora1, lora2, lora3, lora4 = loras
        (
            lora0_checkpoint,
            lora0_weight,
            lora0_alpha,
            lora0_url,
            lora0_file,
        ) = (
            lora0.checkpoint,
            lora0.weight,
            lora0.alpha,
            lora0.url,
            lora0.file,
        )
        (
            lora1_checkpoint,
            lora1_weight,
            lora1_alpha,
            lora1_url,
            lora1_file,
        ) = (
            lora1.checkpoint,
            lora1.weight,
            lora1.alpha,
            lora1.url,
            lora1.file,
        )
        (
            lora2_checkpoint,
            lora2_weight,
            lora2_alpha,
            lora2_url,
            lora2_file,
        ) = (
            lora2.checkpoint,
            lora2.weight,
            lora2.alpha,
            lora2.url,
            lora2.file,
        )
        (
            lora3_checkpoint,
            lora3_weight,
            lora3_alpha,
            lora3_url,
            lora3_file,
        ) = (
            lora3.checkpoint,
            lora3.weight,
            lora3.alpha,
            lora3.url,
            lora3.file,
        )
        (
            lora4_checkpoint,
            lora4_weight,
            lora4_alpha,
            lora4_url,
            lora4_file,
        ) = (
            lora4.checkpoint,
            lora4.weight,
            lora4.alpha,
            lora4.url,
            lora4.file,
        )

        generate = create_element("button", "Generate", variant="primary", scale=2)
        output_image = create_element("image", "Output Image")

        # create layouts
        top1 = create_column(pretrain_layout)
        top2 = create_row(
            create_column(prompt, negative_prompt, scale=4),
            create_column(generate),
        )
        left_generation = create_tab(
            create_row(scheduler, steps),
            create_row(height),
            create_row(width),
            create_row(guidance_scale),
            create_row(seed),
            create_row(freeu_layout),
            name="Generation",
        )
        left_extension = create_tab(
            create_row(controlnet_layout),
            create_row(adapter_layout),
            create_row(lora_layout),
            name="Extensions",
        )
        left_settings = create_tab(
            name="Settings",
        )

        left = create_tabs(left_generation, left_extension, left_settings)
        right = create_column(output_image)
        iface = create_blocks(top1, top2, create_row(left, right))

        # create events
        iface.__enter__()

        start.click(fn=self.start, inputs=[name], outputs=[status])
        stop.click(fn=self.stop, outputs=[status])

        controlnet0_input_image.upload(
            fn=self.processing_controlnet_inputs,
            inputs=[controlnet0_input_image, controlnet0_process],
            outputs=[controlnet0_output_image],
        )
        controlnet0_process.change(
            fn=self.processing_controlnet_inputs,
            inputs=[controlnet0_input_image, controlnet0_process],
            outputs=[controlnet0_output_image],
        )
        controlnet1_input_image.upload(
            fn=self.processing_controlnet_inputs,
            inputs=[controlnet1_input_image, controlnet1_process],
            outputs=[controlnet1_output_image],
        )
        controlnet1_process.change(
            fn=self.processing_controlnet_inputs,
            inputs=[controlnet1_input_image, controlnet1_process],
            outputs=[controlnet1_output_image],
        )
        controlnet2_input_image.upload(
            fn=self.processing_controlnet_inputs,
            inputs=[controlnet2_input_image, controlnet2_process],
            outputs=[controlnet2_output_image],
        )
        controlnet2_process.change(
            fn=self.processing_controlnet_inputs,
            inputs=[controlnet2_input_image, controlnet2_process],
            outputs=[controlnet2_output_image],
        )
        controlnet3_input_image.upload(
            fn=self.processing_controlnet_inputs,
            inputs=[controlnet3_input_image, controlnet3_process],
            outputs=[controlnet3_output_image],
        )
        controlnet3_process.change(
            fn=self.processing_controlnet_inputs,
            inputs=[controlnet3_input_image, controlnet3_process],
            outputs=[controlnet3_output_image],
        )
        controlnet4_input_image.upload(
            fn=self.processing_controlnet_inputs,
            inputs=[controlnet4_input_image, controlnet4_process],
            outputs=[controlnet4_output_image],
        )
        controlnet4_process.change(
            fn=self.processing_controlnet_inputs,
            inputs=[controlnet4_input_image, controlnet4_process],
            outputs=[controlnet4_output_image],
        )

        adapter0_input_image.upload(
            fn=self.processing_adapter_inputs,
            inputs=[adapter0_input_image, adapter0_process],
            outputs=[adapter0_output_image],
        )
        adapter0_process.change(
            fn=self.processing_adapter_inputs,
            inputs=[adapter0_input_image, adapter0_process],
            outputs=[adapter0_output_image],
        )
        adapter1_input_image.upload(
            fn=self.processing_adapter_inputs,
            inputs=[adapter1_input_image, adapter1_process],
            outputs=[adapter1_output_image],
        )
        adapter1_process.change(
            fn=self.processing_adapter_inputs,
            inputs=[adapter1_input_image, adapter1_process],
            outputs=[adapter1_output_image],
        )
        adapter2_input_image.upload(
            fn=self.processing_adapter_inputs,
            inputs=[adapter2_input_image, adapter2_process],
            outputs=[adapter2_output_image],
        )
        adapter2_process.change(
            fn=self.processing_adapter_inputs,
            inputs=[adapter2_input_image, adapter2_process],
            outputs=[adapter2_output_image],
        )
        adapter3_input_image.upload(
            fn=self.processing_adapter_inputs,
            inputs=[adapter3_input_image, adapter3_process],
            outputs=[adapter3_output_image],
        )
        adapter3_process.change(
            fn=self.processing_adapter_inputs,
            inputs=[adapter3_input_image, adapter3_process],
            outputs=[adapter3_output_image],
        )
        adapter4_input_image.upload(
            fn=self.processing_adapter_inputs,
            inputs=[adapter4_input_image, adapter4_process],
            outputs=[adapter4_output_image],
        )
        adapter4_process.change(
            fn=self.processing_adapter_inputs,
            inputs=[adapter4_input_image, adapter4_process],
            outputs=[adapter4_output_image],
        )

        generate.click(
            fn=self.serve,
            inputs=[
                prompt,
                negative_prompt,
                height,
                width,
                guidance_scale,
                steps,
                seed,
                scheduler,
                s1,
                s2,
                b1,
                b2,
                controlnet0_checkpoint,
                controlnet0_output_image,
                controlnet0_guidance_scale,
                controlnet1_checkpoint,
                controlnet1_output_image,
                controlnet1_guidance_scale,
                controlnet2_checkpoint,
                controlnet2_output_image,
                controlnet2_guidance_scale,
                controlnet3_checkpoint,
                controlnet3_output_image,
                controlnet3_guidance_scale,
                controlnet4_checkpoint,
                controlnet4_output_image,
                controlnet4_guidance_scale,
                adapter0_checkpoint,
                adapter0_output_image,
                adapter0_guidance_scale,
                adapter1_checkpoint,
                adapter1_output_image,
                adapter1_guidance_scale,
                adapter2_checkpoint,
                adapter2_output_image,
                adapter2_guidance_scale,
                adapter3_checkpoint,
                adapter3_output_image,
                adapter3_guidance_scale,
                adapter4_checkpoint,
                adapter4_output_image,
                adapter4_guidance_scale,
                lora0_checkpoint,
                lora0_weight,
                lora0_alpha,
                lora0_url,
                lora0_file,
                lora1_checkpoint,
                lora1_weight,
                lora1_alpha,
                lora1_url,
                lora1_file,
                lora2_checkpoint,
                lora2_weight,
                lora2_alpha,
                lora2_url,
                lora2_file,
                lora3_checkpoint,
                lora3_weight,
                lora3_alpha,
                lora3_url,
                lora3_file,
                lora4_checkpoint,
                lora4_weight,
                lora4_alpha,
                lora4_url,
                lora4_file,
            ],
            outputs=[output_image],
        )
        iface.load(
            fn=lambda: [gr.update(value=self._name), gr.update(value=self._status)],
            outputs=[name, status],
        )

        iface.__exit__()

        super().__init__(config, iname="Text2Image", iface=iface)

    def start(self, pretrained_name, **kwargs):
        if self._status == "Running":
            self.stop()
        self._name = pretrained_name
        self._pipe = StableXLForText2ImageGenerationPipeline.from_core_configure(
            self._config,
            pretrained_name=pretrained_name,
        )
        self._status = "Running"
        return self._status

    def stop(self, **kwargs):
        self._pipe.to("cpu")
        del self._pipe
        gc.collect()
        torch.cuda.empty_cache()
        self._pipe = None if not hasattr(self, "_pipe") else self._pipe
        self._status = "Stopped" if self._pipe is None else "Running"
        return self._status

    def processing_controlnet_inputs(self, image, process):
        pfunc = controlnet_processes.get(process, None)
        if pfunc is not None and image is not None:
            return pfunc(image)
        return image

    def processing_adapter_inputs(self, image, process):
        pfunc = adapter_processes.get(process, None)
        if pfunc is not None and image is not None:
            return pfunc(image)
        return image

    def serve(
        self,
        text: str,
        negative_text: Optional[str] = None,
        height: Optional[int] = 512,
        width: Optional[int] = 512,
        guidance_scale: Optional[float] = 7.5,
        num_timesteps: Optional[int] = 25,
        seed: Optional[int] = 1123,
        scheduler: Optional[str] = None,
        freeu_s1: Optional[float] = 0.9,
        freeu_s2: Optional[float] = 0.2,
        freeu_b1: Optional[float] = 1.2,
        freeu_b2: Optional[float] = 1.4,
        controlnet0_checkpoint: Optional[str] = None,
        controlnet0_image: Optional[Image.Image] = None,
        controlnet0_guidance_scale: Optional[float] = 0.8,
        controlnet1_checkpoint: Optional[str] = None,
        controlnet1_image: Optional[Image.Image] = None,
        controlnet1_guidance_scale: Optional[float] = 0.8,
        controlnet2_checkpoint: Optional[str] = None,
        controlnet2_image: Optional[Image.Image] = None,
        controlnet2_guidance_scale: Optional[float] = 0.8,
        controlnet3_checkpoint: Optional[str] = None,
        controlnet3_image: Optional[Image.Image] = None,
        controlnet3_guidance_scale: Optional[float] = 0.8,
        controlnet4_checkpoint: Optional[str] = None,
        controlnet4_image: Optional[Image.Image] = None,
        controlnet4_guidance_scale: Optional[float] = 0.8,
        adapter0_checkpoint: Optional[str] = None,
        adapter0_image: Optional[Image.Image] = None,
        adapter0_guidance_scale: Optional[float] = 1.0,
        adapter1_checkpoint: Optional[str] = None,
        adapter1_image: Optional[Image.Image] = None,
        adapter1_guidance_scale: Optional[float] = 1.0,
        adapter2_checkpoint: Optional[str] = None,
        adapter2_image: Optional[Image.Image] = None,
        adapter2_guidance_scale: Optional[float] = 1.0,
        adapter3_checkpoint: Optional[str] = None,
        adapter3_image: Optional[Image.Image] = None,
        adapter3_guidance_scale: Optional[float] = 1.0,
        adapter4_checkpoint: Optional[str] = None,
        adapter4_image: Optional[Image.Image] = None,
        adapter4_guidance_scale: Optional[float] = 1.0,
        lora0_checkpoint: Optional[str] = None,
        lora0_weight: Optional[float] = 1.0,
        lora0_alpha: Optional[float] = 32,
        lora0_url: Optional[str] = None,
        lora0_file: Optional[str] = None,
        lora1_checkpoint: Optional[str] = None,
        lora1_weight: Optional[float] = 1.0,
        lora1_alpha: Optional[float] = 32,
        lora1_url: Optional[str] = None,
        lora1_file: Optional[str] = None,
        lora2_checkpoint: Optional[str] = None,
        lora2_weight: Optional[float] = 1.0,
        lora2_alpha: Optional[float] = 32,
        lora2_url: Optional[str] = None,
        lora2_file: Optional[str] = None,
        lora3_checkpoint: Optional[str] = None,
        lora3_weight: Optional[float] = 1.0,
        lora3_alpha: Optional[float] = 32,
        lora3_url: Optional[str] = None,
        lora3_file: Optional[str] = None,
        lora4_checkpoint: Optional[str] = None,
        lora4_weight: Optional[float] = 1.0,
        lora4_alpha: Optional[float] = 32,
        lora4_url: Optional[str] = None,
        lora4_file: Optional[str] = None,
    ):
        assert self._pipe is not None
        image = self._pipe(
            text,
            negative_text,
            height=height,
            width=width,
            guidance_scale=guidance_scale,
            num_timesteps=num_timesteps,
            seed=seed,
            scheduler=scheduler,
            freeu_params=(freeu_s1, freeu_s2, freeu_b1, freeu_b2),
            controlnet_checkpoints=(
                controlnet0_checkpoint,
                controlnet1_checkpoint,
                controlnet2_checkpoint,
                controlnet3_checkpoint,
                controlnet4_checkpoint,
            ),
            controlnet_images=(
                controlnet0_image,
                controlnet1_image,
                controlnet2_image,
                controlnet3_image,
                controlnet4_image,
            ),
            controlnet_guidance_scales=(
                controlnet0_guidance_scale,
                controlnet1_guidance_scale,
                controlnet2_guidance_scale,
                controlnet3_guidance_scale,
                controlnet4_guidance_scale,
            ),
            adapter_checkpoints=(
                adapter0_checkpoint,
                adapter1_checkpoint,
                adapter2_checkpoint,
                adapter3_checkpoint,
                adapter4_checkpoint,
            ),
            adapter_images=(
                adapter0_image,
                adapter1_image,
                adapter2_image,
                adapter3_image,
                adapter4_image,
            ),
            adapter_guidance_scales=(
                adapter0_guidance_scale,
                adapter1_guidance_scale,
                adapter2_guidance_scale,
                adapter3_guidance_scale,
                adapter4_guidance_scale,
            ),
            lora_checkpoints=(
                lora0_checkpoint,
                lora1_checkpoint,
                lora2_checkpoint,
                lora3_checkpoint,
                lora4_checkpoint,
            ),
            lora_weights=(
                lora0_weight,
                lora1_weight,
                lora2_weight,
                lora3_weight,
                lora4_weight,
            ),
            lora_alphas=(
                lora0_alpha,
                lora1_alpha,
                lora2_alpha,
                lora3_alpha,
                lora4_alpha,
            ),
            lora_urls=(
                lora0_url,
                lora1_url,
                lora2_url,
                lora3_url,
                lora4_url,
            ),
            lora_files=(
                lora0_file,
                lora1_file,
                lora2_file,
                lora3_file,
                lora4_file,
            ),
        )
        return image



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.models.llama.modeling import (
    LlamaForClassification,
    LlamaForGeneration,
)
from unitorch.models.llama.processing import LlamaProcessor



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import torch
import json
import numpy as np
from PIL import Image
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from transformers import CLIPTokenizer, CLIPImageProcessor
from torchvision.transforms import (
    Resize,
    CenterCrop,
    RandomCrop,
    Lambda,
    ToTensor,
    Normalize,
    Compose,
    RandomHorizontalFlip,
)
from diffusers.image_processor import VaeImageProcessor
from unitorch.models import (
    HfTextClassificationProcessor,
    HfImageClassificationProcessor,
    GenericOutputs,
)


class StableProcessor(HfTextClassificationProcessor):
    def __init__(
        self,
        vocab_path: str,
        merge_path: str,
        vae_config_path: Optional[str] = None,
        max_seq_length: Optional[int] = 77,
        position_start_id: Optional[int] = 0,
        pad_token: Optional[str] = "<|endoftext|>",
        image_size: Optional[Tuple[int, int]] = None,
        center_crop: Optional[bool] = False,
        random_flip: Optional[bool] = False,
    ):
        tokenizer = CLIPTokenizer(
            vocab_file=vocab_path,
            merges_file=merge_path,
        )

        tokenizer.cls_token = tokenizer.bos_token
        tokenizer.sep_token = tokenizer.eos_token
        tokenizer.pad_token = pad_token

        HfTextClassificationProcessor.__init__(
            self,
            tokenizer=tokenizer,
            max_seq_length=max_seq_length,
            position_start_id=position_start_id,
        )

        self.image_size = image_size

        if self.image_size is not None:
            self.vision_processor = Compose(
                [
                    Resize(self.image_size),
                    CenterCrop(self.image_size)
                    if center_crop
                    else RandomCrop(self.image_size),
                    RandomHorizontalFlip() if random_flip else Lambda(lambda x: x),
                    ToTensor(),
                    Normalize([0.5], [0.5]),
                ]
            )
            self.condition_vision_processor = Compose(
                [
                    Resize(self.image_size),
                    CenterCrop(self.image_size),
                    ToTensor(),
                ]
            )
        else:
            self.vision_processor = Compose(
                [
                    RandomHorizontalFlip() if random_flip else Lambda(lambda x: x),
                    ToTensor(),
                    Normalize([0.5], [0.5]),
                ]
            )
            self.condition_vision_processor = Compose(
                [
                    ToTensor(),
                ]
            )

        if vae_config_path is not None:
            vae_config_dict = json.load(open(vae_config_path))
            vae_scale_factor = 2 ** (
                len(vae_config_dict.get("block_out_channels", [])) - 1
            )
            self.vae_image_processor = VaeImageProcessor(
                vae_scale_factor=vae_scale_factor
            )
            self.vae_mask_image_processor = VaeImageProcessor(
                vae_scale_factor=vae_scale_factor
            )
            self.vae_condition_image_processor = VaeImageProcessor(
                vae_scale_factor=vae_scale_factor,
                do_convert_rgb=True,
                do_normalize=False,
            )
        else:
            self.vae_image_processor = None
            self.vae_mask_image_processor = None
            self.vae_condition_image_processor = None

    def text2image(
        self,
        prompt: str,
        image: Union[Image.Image, str],
        max_seq_length: Optional[int] = None,
    ):
        if isinstance(image, str):
            image = Image.open(image)

        image = image.convert("RGB")

        pixel_values = self.vision_processor(image)

        prompt_outputs = self.classification(prompt, max_seq_length=max_seq_length)

        return GenericOutputs(
            pixel_values=pixel_values,
            input_ids=prompt_outputs.input_ids,
            attention_mask=prompt_outputs.attention_mask,
        )

    def text2image_inputs(
        self,
        prompt: str,
        negative_prompt: Optional[str] = "",
        max_seq_length: Optional[int] = None,
    ):
        prompt_outputs = self.classification(prompt, max_seq_length=max_seq_length)
        negative_prompt_outputs = self.classification(
            negative_prompt, max_seq_length=max_seq_length
        )

        return GenericOutputs(
            input_ids=prompt_outputs.input_ids,
            attention_mask=prompt_outputs.attention_mask,
            negative_input_ids=negative_prompt_outputs.input_ids,
            negative_attention_mask=negative_prompt_outputs.attention_mask,
        )

    def image2image_inputs(
        self,
        image: Union[Image.Image, str],
    ):
        if isinstance(image, str):
            image = Image.open(image)
        image = image.convert("RGB")
        if self.image_size is not None:
            image = image.resize(self.image_size)

        pixel_values = self.vae_image_processor.preprocess(image)[0]

        return GenericOutputs(pixel_values=pixel_values)

    def inpainting_inputs(
        self,
        image: Union[Image.Image, str],
        mask_image: Union[Image.Image, str],
    ):
        if isinstance(image, str):
            image = Image.open(image)
        image = image.convert("RGB")

        if isinstance(mask_image, str):
            mask_image = Image.open(mask_image)
        mask_image = mask_image.convert("L")

        if self.image_size is not None:
            image = image.resize(self.image_size)
            mask_image = mask_image.resize(self.image_size)

        pixel_values = self.vae_image_processor.preprocess(image)[0]
        pixel_masks = self.vae_image_processor.preprocess(mask_image)[0]
        pixel_masks = (pixel_masks + 1) / 2

        return GenericOutputs(
            pixel_values=pixel_values,
            pixel_masks=pixel_masks,
        )

    def resolution_inputs(
        self,
        image: Union[Image.Image, str],
    ):
        if isinstance(image, str):
            image = Image.open(image)
        image = image.convert("RGB")
        if self.image_size is not None:
            image = image.resize(self.image_size)

        pixel_values = self.vae_image_processor.preprocess(image)[0]

        return GenericOutputs(pixel_values=pixel_values)

    def controlnet_inputs(
        self,
        image: Union[Image.Image, str],
    ):
        if isinstance(image, str):
            image = Image.open(image)
        image = image.convert("RGB")
        if self.image_size is not None:
            image = image.resize(self.image_size)

        pixel_values = self.vae_condition_image_processor.preprocess(image)[0]
        return GenericOutputs(pixel_values=pixel_values)

    def controlnets_inputs(
        self,
        images: List[Union[Image.Image, str]],
    ):
        pixel_values = []
        for image in images:
            if isinstance(image, str):
                image = Image.open(image)
            image = image.convert("RGB")
            if self.image_size is not None:
                image = image.resize(self.image_size)

            pixel_values.append(self.vae_condition_image_processor.preprocess(image)[0])

        return GenericOutputs(pixel_values=torch.stack(pixel_values, dim=0))

    def adapter_inputs(
        self,
        image: Union[Image.Image, str],
    ):
        if isinstance(image, str):
            image = Image.open(image)
        image = image.convert("RGB")
        if self.image_size is not None:
            image = image.resize(self.image_size)

        pixel_values = self.vae_image_processor.preprocess(image)[0]

        return GenericOutputs(pixel_values=pixel_values)

    def adapters_inputs(
        self,
        images: List[Union[Image.Image, str]],
    ):
        pixel_values = []
        for image in images:
            if isinstance(image, str):
                image = Image.open(image)
            image = image.convert("RGB")
            if self.image_size is not None:
                image = image.resize(self.image_size)

            pixel_values.append(self.vae_image_processor.preprocess(image)[0])

        return GenericOutputs(pixel_values=torch.stack(pixel_values, dim=0))


class StableVideoProcessor(HfImageClassificationProcessor):
    def __init__(
        self,
        vision_config_path: str,
        vae_config_path: str,
    ):
        vision_processor = CLIPImageProcessor.from_json_file(vision_config_path)
        super().__init__(
            vision_processor=vision_processor,
        )
        vae_config_dict = json.load(open(vae_config_path))
        vae_scale_factor = 2 ** (len(vae_config_dict.get("block_out_channels", [])) - 1)
        self.vae_image_processor = VaeImageProcessor(vae_scale_factor=vae_scale_factor)

    def image2video_inputs(
        self,
        image: Union[Image.Image, str],
        vae_image: Union[Image.Image, str],
    ):
        if isinstance(image, str):
            image = Image.open(image)
        image = image.resize((224, 224), Image.BICUBIC)
        pixel_outputs = self.classification(image=image)
        if isinstance(vae_image, str):
            vae_image = Image.open(vae_image)
        vae_pixel_values = self.vae_image_processor.preprocess(vae_image)[0]

        return GenericOutputs(
            pixel_values=pixel_outputs.pixel_values,
            vae_pixel_values=vae_pixel_values,
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

# pretrained infos
pretrained_xpegasus_infos = {
    "default-xpegasus": {
        "config": "https://huggingface.co/google/pegasus-x-base/resolve/main/config.json",
        "vocab": "https://huggingface.co/google/pegasus-x-base/resolve/main/spiece.model",
    },
    "xpegasus-base": {
        "config": "https://huggingface.co/google/pegasus-x-base/resolve/main/config.json",
        "vocab": "https://huggingface.co/google/pegasus-x-base/resolve/main/spiece.model",
        "weight": "https://huggingface.co/google/pegasus-x-base/resolve/main/pytorch_model.bin",
    },
    "xpegasus-large": {
        "config": "https://huggingface.co/google/pegasus-x-large/resolve/main/config.json",
        "vocab": "https://huggingface.co/google/pegasus-x-large/resolve/main/spiece.model",
        "weight": "https://huggingface.co/google/pegasus-x-large/resolve/main/pytorch_model.bin",
    },
}

import unitorch.cli.models.xpegasus.modeling
import unitorch.cli.models.xpegasus.processing
from unitorch.cli.models.xpegasus.modeling import XPegasusForGeneration
from unitorch.cli.models.xpegasus.processing import XPegasusProcessor



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import torch
import torch.nn as nn
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from unitorch.losses.ranking import ListMLELoss, ApproxNDCGLoss, ApproxMRRLoss


class CELoss(nn.Module):
    def __init__(
        self,
        smoothing_alpha: Optional[float] = 0.0,
        weight: Optional[torch.Tensor] = None,
        reduction: Optional[str] = "mean",
    ):
        super().__init__()
        self.smoothing_alpha = smoothing_alpha
        self.weight = weight
        self.reduction = reduction

    def forward(
        self,
        input: torch.Tensor,
        target: torch.Tensor,
        sample_weight: Optional[torch.Tensor] = None,
    ):
        assert input.dim() == 2 and target.dim() == 1

        target = target.long()

        if self.smoothing_alpha > 0:
            batch_size, num_classes = input.size()
            smooth_label = torch.full(
                size=(batch_size, num_classes),
                fill_value=self.smoothing_alpha / (num_classes - 1),
            ).to(loss)
            smooth_label.scatter_(
                dim=1,
                index=torch.unsqueeze(target, dim=1),
                value=(1 - self.smoothing_alpha),
            )
            log_logits = torch.nn.functional.log_softmax(input, dim=1)
            loss = torch.sum(log_logits * smooth_label, dim=1)
        else:
            loss = nn.CrossEntropyLoss(weight=self.weight, reduction="none")(
                input, target
            ).squeeze()

        if sample_weight is not None:
            loss = loss * sample_weight

        if self.reduction == "mean":
            loss = torch.mean(loss)

        return loss


class BCELoss(nn.Module):
    def __init__(
        self,
        weight: Optional[torch.Tensor] = None,
        reduction: Optional[str] = "mean",
    ):
        super().__init__()
        self.weight = weight
        self.reduction = reduction

    def forward(
        self,
        input: torch.Tensor,
        target: torch.Tensor,
        sample_weight: Optional[torch.Tensor] = None,
    ):
        if target.dim() == 1:
            target = torch.unsqueeze(target, dim=1)
        assert input.dim() == 2 and target.dim() == 2

        target = target.float()
        loss = nn.BCEWithLogitsLoss(weight=self.weight, reduction="none")(input, target)

        loss = torch.sum(loss, dim=1)

        if sample_weight is not None:
            loss = loss * sample_weight

        if self.reduction == "mean":
            loss = torch.mean(loss)

        return loss


class LMLoss(nn.Module):
    def __init__(
        self,
        reduction: Optional[str] = "mean",
    ):
        super().__init__()
        self.reduction = reduction

    def forward(
        self,
        input: torch.Tensor,
        target: torch.Tensor,
        masks: Optional[torch.Tensor] = None,
        sample_weight: Optional[torch.Tensor] = None,
    ):
        assert input.dim() == 3 and target.dim() == 2

        batch_size, seq_len, num_classes = input.size()
        input = input.contiguous().view(batch_size * seq_len, num_classes)
        target = target.contiguous().view(-1)
        target = target.long()
        if masks is None:
            masks = torch.ones_like(target).to(target)
        masks = masks.contiguous().view(-1)
        loss = nn.CrossEntropyLoss(reduction="none")(input, target)
        loss = loss * masks.float()
        loss = loss.contiguous().view(batch_size, seq_len).sum(1) / torch.max(
            masks.contiguous().view(batch_size, seq_len).float().sum(1),
            torch.ones(batch_size).to(masks.device),
        )

        if sample_weight is not None:
            loss = loss * sample_weight

        if self.reduction == "mean":
            loss = torch.mean(loss)

        return loss


class MSELoss(nn.Module):
    def __init__(
        self,
        reduction: Optional[str] = "mean",
    ):
        super().__init__()
        self.reduction = reduction

    def forward(
        self,
        input: torch.Tensor,
        target: torch.Tensor,
        sample_weight: Optional[torch.Tensor] = None,
    ):
        assert input.size(0) == target.size(0) and input.numel() == target.numel()
        batch_size = input.size(0)
        loss = nn.MSELoss(reduction="none")(
            input.view(batch_size, -1), target.view(batch_size, -1)
        )

        if loss.dim() > 1:
            loss = torch.sum(loss, dim=1)

        if sample_weight is not None:
            loss = loss * sample_weight

        if self.reduction == "mean":
            loss = torch.mean(loss)

        return loss



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import io
import requests
import time
import base64
import json
import logging
import torch
import torch.nn as nn
import numpy as np
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from random import random
from PIL import Image, ImageOps, ImageFile, ImageFilter

from unitorch.cli import (
    add_default_section_for_init,
    add_default_section_for_function,
    register_process,
)

ImageFile.LOAD_TRUNCATED_IMAGES = True


class ImageProcessor:
    """
    Processor for image-related operations.
    """

    def __init__(
        self,
        image_type: Optional[str] = None,
        image_size: tuple = (256, 256),
        http_url: Optional[str] = None,
    ):
        """
        Initializes a new instance of the ImageProcessor.

        Args:
            image_type (Optional[str]): The type of the image. Defaults to None.
            image_size (tuple): The size of the image. Defaults to (256, 256).
            http_url (Optional[str]): The URL for fetching images. Defaults to None.
        """
        self.image_type = image_type
        self.image_size = image_size
        self.http_url = http_url

    @classmethod
    @add_default_section_for_init("core/process/image")
    def from_core_configure(cls, config, **kwargs):
        """
        Creates a new instance of the ImageProcessor using the configuration from the core.

        Args:
            config: The configuration object.
            **kwargs: Additional keyword arguments.

        Returns:
            An instance of the ImageProcessor.
        """
        pass

    def _request_url(self, url):
        """
        Sends an HTTP request to the specified URL and returns the response.

        Args:
            url (str): The URL to request.

        Returns:
            The response received from the URL.
        """
        while True:
            try:
                doc = requests.get(url, timeout=600)
                return doc
            except:
                time.sleep(random() * 2)

    @register_process("core/process/image/read")
    def _read(
        self,
        image,
        image_type=None,
    ):
        """
        Reads and processes an image.

        Args:
            image: The image to read and process.
            image_type (Optional[str]): The type of the image. Defaults to None.

        Returns:
            The processed image as a PIL Image object.
        """
        image_type = image_type if image_type is not None else self.image_type
        try:
            if image_type == "base64":
                image = io.BytesIO(base64.b64decode(image))
                return Image.open(image).convert("RGB")

            if image_type == "hex":
                image = io.BytesIO(bytes.fromhex(image))
                return Image.open(image).convert("RGB")

            if self.http_url is None:
                return Image.open(image).convert("RGB")

            url = self.http_url.format(image)
            doc = self._request_url(url)
            if doc.status_code != 200 or doc.content == b"":
                raise ValueError(f"can't find the image {image}")

            return Image.open(io.BytesIO(doc.content)).convert("RGB")
        except:
            logging.debug(f"core/process/image/read use fake image for {image}")
            return Image.new("RGB", self.image_size, (255, 255, 255))

    @register_process("core/process/image/translate")
    def _translate(
        self,
        image: Image.Image,
        up: Optional[int] = 0,
        left: Optional[int] = 0,
    ):
        """
        Translates the image by the specified amount.

        Args:
            image (Image.Image): The image to translate.
            up (Optional[int]): The amount to move the image upwards. Defaults to 0.
            left (Optional[int]): The amount to move the image to the left. Defaults to 0.

        Returns:
            The translated image as a PIL Image object.
        """
        return image.transform(image.size, Image.AFFINE, (1, 0, left, 0, 1, up))

    @register_process("core/process/image/scale")
    def _scale(
        self,
        image: Image.Image,
        scale: Optional[float] = 0.0,
    ):
        """
        Scales the image by the specified factor.

        Args:
            image (Image.Image): The image to scale.
            scale (Optional[float]): The scale factor. Defaults to 0.0.

        Returns:
            The scaled image as a PIL Image object.
        """
        return ImageOps.scale(image, scale)

    @register_process("core/process/image/rotate")
    def _rotate(
        self,
        image: Image.Image,
        degree: Optional[int] = 0.0,
    ):
        """
        Rotates the image by the specified degree.

        Args:
            image (Image.Image): The image to rotate.
            degree (Optional[int]): The degree of rotation. Defaults to 0.0.

        Returns:
            The rotated image as a PIL Image object.
        """
        return image.rotate(degree, Image.NEAREST, expand=1)

    @register_process("core/process/image/flip")
    def _flip(
        self,
        image: Image.Image,
        horizontal: Optional[bool] = False,
        vertical: Optional[bool] = False,
    ):
        """
        Flips the image horizontally or vertically.

        Args:
            image (Image.Image): The image to flip.
            horizontal (Optional[bool]): Whether to flip the image horizontally. Defaults to False.
            vertical (Optional[bool]): Whether to flip the image vertically. Defaults to False.

        Returns:
            The flipped image as a PIL Image object.
        """
        if horizontal:
            image = image.transpose(Image.FLIP_LEFT_RIGHT)
        if vertical:
            image = image.transpose(Image.FLIP_TOP_BOTTOM)
        return image

    @register_process("core/process/image/resize")
    def _resize(
        self,
        image: Image.Image,
        size: Optional[Tuple[int, int]] = (256, 256),
    ):
        """
        Resizes the image to the specified size.

        Args:
            image (Image.Image): The image to resize.
            size (Optional[Tuple[int, int]]): The size to resize to. Defaults to (256, 256).

        Returns:
            The resized image as a PIL Image object.
        """
        return image.resize(size, Image.BICUBIC)

    @register_process("core/process/image/canny")
    def _canny(
        self,
        image: Image.Image,
    ):
        """
        Detects edges in the image using the Canny algorithm.

        Args:
            image (Image.Image): The image to detect edges in.

        Returns:
            The image with detected edges as a PIL Image object.
        """
        return image.filter(ImageFilter.FIND_EDGES)



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
from datasets import load_dataset, Dataset
from torch.utils.data import Dataset as TorchDataset, IterableDataset
from datasets.features import Features
from itertools import cycle
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union


class HFDatasets(TorchDataset):
    def __init__(self, dataset: Dataset):
        self.dataset = dataset

    @classmethod
    def from_csv(
        cls,
        data_dir: Optional[str] = None,
        data_files: Optional[Union[str, List[str]]] = None,
        names: Optional[List[str]] = None,
        sep: Optional[str] = "\t",
        quoting: Optional[int] = 3,
        escapechar: Optional[str] = None,
        split: Optional[str] = None,
    ) -> Optional["HFDatasets"]:
        if data_files is None:
            return None

        if isinstance(data_files, str) and not os.path.exists(data_files):
            return None

        dataset = load_dataset(
            "csv",
            data_dir=data_dir,
            data_files=data_files,
            delimiter=sep,
            column_names=names,
            quoting=quoting,
            escapechar=escapechar,
        )

        if split not in dataset:
            dataset = dataset.get("train")
        else:
            dataset = dataset.get(split)

        return cls(dataset)

    @classmethod
    def from_json(
        cls,
        data_dir: Optional[str] = None,
        data_files: Optional[Union[str, List[str]]] = None,
        field: Optional[str] = None,
        split: Optional[str] = None,
    ) -> Optional["HFDatasets"]:
        if data_files is None:
            return None

        if isinstance(data_files, str) and not os.path.exists(data_files):
            return None

        dataset = load_dataset(
            "json",
            data_dir=data_dir,
            data_files=data_files,
            field=field,
        )

        if split not in dataset:
            dataset = dataset.get("train")
        else:
            dataset = dataset.get(split)

        return cls(dataset)

    @classmethod
    def from_parquet(
        cls,
        data_dir: Optional[str] = None,
        data_files: Optional[Union[str, List[str]]] = None,
        split: Optional[str] = None,
        features: Optional[Features] = None,
    ) -> Optional["HFDatasets"]:
        if data_files is None:
            return None

        if isinstance(data_files, str) and not os.path.exists(data_files):
            return None

        dataset = load_dataset(
            "parquet",
            data_dir=data_dir,
            data_files=data_files,
            features=features,
        )

        if split not in dataset:
            dataset = dataset.get("train")
        else:
            dataset = dataset.get(split)

        return cls(dataset)

    @classmethod
    def from_hub(
        cls,
        data_name: str,
        config_name: Optional[str] = None,
        data_dir: Optional[str] = None,
        data_files: Optional[Union[str, List[str]]] = None,
        split: Optional[str] = None,
    ) -> Optional["HFDatasets"]:
        dataset = load_dataset(
            data_name,
            name=config_name,
            data_dir=data_dir,
            data_files=data_files,
        )

        if split in dataset:
            dataset = dataset.get(split)
            return cls(dataset)

    def __getitem__(self, idx: int) -> Any:
        return self.dataset[idx]

    def __len__(self) -> int:
        return len(self.dataset)


class HFIterableDatasets(IterableDataset):
    def __init__(
        self,
        dataset: Dataset,
    ):
        self.dataset = dataset

    def set_epoch(self, epoch: int) -> None:
        self.dataset.set_epoch(epoch)

    @classmethod
    def from_csv(
        cls,
        data_dir: Optional[str] = None,
        data_files: Optional[Union[str, List[str]]] = None,
        names: Optional[List[str]] = None,
        sep: Optional[str] = "\t",
        split: Optional[str] = None,
    ) -> Optional["HFIterableDatasets"]:
        if data_files is None:
            return None

        if isinstance(data_files, str) and not os.path.exists(data_files):
            return None

        dataset = load_dataset(
            "csv",
            data_dir=data_dir,
            data_files=data_files,
            delimiter=sep,
            column_names=names,
            quoting=3,
            streaming=True,
        )

        if split not in dataset:
            dataset = dataset.get("train")
        else:
            dataset = dataset.get(split)

        return cls(dataset)

    @classmethod
    def from_json(
        cls,
        data_dir: Optional[str] = None,
        data_files: Optional[Union[str, List[str]]] = None,
        field: Optional[str] = None,
        split: Optional[str] = None,
    ) -> Optional["HFIterableDatasets"]:
        if data_files is None:
            return None

        if isinstance(data_files, str) and not os.path.exists(data_files):
            return None

        dataset = load_dataset(
            "json",
            data_dir=data_dir,
            data_files=data_files,
            field=field,
            streaming=True,
        )

        if split not in dataset:
            dataset = dataset.get("train")
        else:
            dataset = dataset.get(split)

        return cls(dataset)

    @classmethod
    def from_parquet(
        cls,
        data_dir: Optional[str] = None,
        data_files: Optional[Union[str, List[str]]] = None,
        split: Optional[str] = None,
    ) -> Optional["HFIterableDatasets"]:
        if data_files is None:
            return None

        if isinstance(data_files, str) and not os.path.exists(data_files):
            return None

        dataset = load_dataset(
            "parquet",
            data_dir=data_dir,
            data_files=data_files,
            streaming=True,
        )

        if split not in dataset:
            dataset = dataset.get("train")
        else:
            dataset = dataset.get(split)

        return cls(dataset)

    @classmethod
    def from_hub(
        cls,
        data_name: str,
        config_name: Optional[str] = None,
        data_dir: Optional[str] = None,
        data_files: Optional[Union[str, List[str]]] = None,
        split: Optional[str] = None,
    ) -> Optional["HFIterableDatasets"]:
        dataset = load_dataset(
            data_name,
            name=config_name,
            data_dir=data_dir,
            data_files=data_files,
            streaming=True,
        )

        if split in dataset:
            dataset = dataset.get(split)
            return cls(dataset)

    def __iter__(self) -> Iterable:
        return iter(cycle(self.dataset))

    def __len__(self) -> int:
        return len(self.dataset)



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

# pretrained infos
pretrained_clip_infos = {
    "default-clip": {
        "config": "https://huggingface.co/openai/clip-vit-base-patch16/resolve/main/config.json",
        "vision_config": "https://huggingface.co/openai/clip-vit-base-patch16/resolve/main/preprocessor_config.json",
        "vocab": "https://huggingface.co/openai/clip-vit-base-patch16/resolve/main/vocab.json",
        "merge": "https://huggingface.co/openai/clip-vit-base-patch16/resolve/main/merges.txt",
    },
    "clip-vit-base-patch16": {
        "config": "https://huggingface.co/openai/clip-vit-base-patch16/resolve/main/config.json",
        "vision_config": "https://huggingface.co/openai/clip-vit-base-patch16/resolve/main/preprocessor_config.json",
        "vocab": "https://huggingface.co/openai/clip-vit-base-patch16/resolve/main/vocab.json",
        "merge": "https://huggingface.co/openai/clip-vit-base-patch16/resolve/main/merges.txt",
        "weight": "https://huggingface.co/openai/clip-vit-base-patch16/resolve/main/pytorch_model.bin",
    },
    "clip-vit-base-patch32": {
        "config": "https://huggingface.co/openai/clip-vit-base-patch32/resolve/main/config.json",
        "vision_config": "https://huggingface.co/openai/clip-vit-base-patch32/resolve/main/preprocessor_config.json",
        "vocab": "https://huggingface.co/openai/clip-vit-base-patch32/resolve/main/vocab.json",
        "merge": "https://huggingface.co/openai/clip-vit-base-patch32/resolve/main/merges.txt",
        "weight": "https://huggingface.co/openai/clip-vit-base-patch32/resolve/main/pytorch_model.bin",
    },
    "clip-vit-large-patch14": {
        "config": "https://huggingface.co/openai/clip-vit-large-patch14/resolve/main/config.json",
        "vision_config": "https://huggingface.co/openai/clip-vit-large-patch14/resolve/main/preprocessor_config.json",
        "vocab": "https://huggingface.co/openai/clip-vit-large-patch14/resolve/main/vocab.json",
        "merge": "https://huggingface.co/openai/clip-vit-large-patch14/resolve/main/merges.txt",
        "weight": "https://huggingface.co/openai/clip-vit-large-patch14/resolve/main/pytorch_model.bin",
    },
}

import unitorch.cli.models.clip.modeling
import unitorch.cli.models.clip.processing
from unitorch.cli.models.clip.modeling import (
    ClipForPretrain,
    ClipForClassification,
    ClipForTextClassification,
    ClipForImageClassification,
)
from unitorch.cli.models.clip.processing import ClipProcessor



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import io
import torch
import gc
import gradio as gr
from PIL import Image
from unitorch.cli import CoreConfigureParser, GenericWebUI
from unitorch.cli import register_webui
from unitorch.cli.pipelines.blip import BlipForImageCaptionPipeline
from unitorch.cli.models.blip import pretrained_blip_infos
from unitorch.cli.webuis import (
    matched_pretrained_names,
    create_element,
    create_accordion,
    create_row,
    create_column,
    create_group,
    create_tab,
    create_tabs,
    create_blocks,
    create_pretrain_layout,
    create_lora_layout,
    create_freeu_layout,
)
from unitorch.cli.webuis import SimpleWebUI


@register_webui("core/webui/blip")
class BlipCaptionWebUI(SimpleWebUI):
    pretrained_names = list(pretrained_blip_infos.keys())
    supported_pretrained_names = matched_pretrained_names(pretrained_names, "^blip-")

    def __init__(self, config: CoreConfigureParser):
        self._pipe = None if not hasattr(self, "_pipe") else self._pipe
        self._status = "Stopped" if self._pipe is None else "Running"
        if len(self.supported_pretrained_names) == 0:
            raise ValueError("No supported pretrained models found.")
        self._name = self.supported_pretrained_names[0]

        # create elements
        pretrain_layout_group = create_pretrain_layout(
            self.supported_pretrained_names, self._name
        )
        name, status, start, stop, pretrain_layout = (
            pretrain_layout_group.name,
            pretrain_layout_group.status,
            pretrain_layout_group.start,
            pretrain_layout_group.stop,
            pretrain_layout_group.layout,
        )

        image = create_element("image", "Input Image")
        generate = create_element("button", "Generate")
        result = create_element("text", "Output Result", lines=3)

        # create blocks
        left = create_column(image, generate)
        right = create_column(result)
        iface = create_blocks(pretrain_layout, create_row(left, right))

        # create events
        iface.__enter__()

        start.click(self.start, inputs=[name], outputs=[status])
        stop.click(self.stop, outputs=[status])
        generate.click(self.serve, inputs=[image], outputs=[result])

        iface.load(
            fn=lambda: [gr.update(value=self._name), gr.update(value=self._status)],
            outputs=[name, status],
        )

        iface.__exit__()

        super().__init__(config, iname="Blip", iface=iface)

    def start(self, pretrained_name, **kwargs):
        if self._status == "Running":
            self.stop()
        self._name = pretrained_name
        self._pipe = BlipForImageCaptionPipeline.from_core_configure(
            self._config,
            pretrained_name=pretrained_name,
        )
        self._status = "Running"
        return self._status

    def stop(self, **kwargs):
        self._pipe.to("cpu")
        del self._pipe
        gc.collect()
        torch.cuda.empty_cache()
        self._pipe = None if not hasattr(self, "_pipe") else self._pipe
        self._status = "Stopped" if self._pipe is None else "Running"
        return self._status

    def serve(
        self,
        image: Image.Image,
    ):
        assert self._pipe is not None
        result = self._pipe(image)
        return result



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import io
import torch
import gc
import gradio as gr
from PIL import Image
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from unitorch.cli import CoreConfigureParser, GenericWebUI
from unitorch.cli import register_webui
from unitorch.cli.models.llama import (
    pretrained_llama_infos,
    pretrained_llama_extensions_infos,
)
from unitorch.cli.pipelines.llama import LlamaForGenerationPipeline
from unitorch.cli.webuis import (
    matched_pretrained_names,
    create_element,
    create_accordion,
    create_row,
    create_column,
    create_group,
    create_tab,
    create_tabs,
    create_blocks,
    create_pretrain_layout,
    create_lora_layout,
    create_freeu_layout,
)
from unitorch.cli.webuis import SimpleWebUI


@register_webui("core/webui/llama")
class LlamaWebUI(SimpleWebUI):
    pretrained_names = list(pretrained_llama_infos.keys())
    supported_pretrained_names = matched_pretrained_names(
        pretrained_names,
        ["^llama-", "^llama2-", "^open-llama-", "^llama3-"],
    )
    pretrained_extension_names = list(pretrained_llama_extensions_infos.keys())
    supported_lora_names = matched_pretrained_names(
        pretrained_extension_names,
        ["^llama-lora-", "^llama2-lora-", "^open-llama-lora-", "^llama3-lora-"],
    )

    def __init__(self, config: CoreConfigureParser):
        self._config = config
        self._pipe = None if not hasattr(self, "_pipe") else self._pipe
        self._status = "Stopped" if self._pipe is None else "Running"
        if len(self.supported_pretrained_names) == 0:
            raise ValueError("No supported pretrained models found.")
        self._name = self.supported_pretrained_names[0]

        # Create the elements
        pretrain_layout_group = create_pretrain_layout(
            self.supported_pretrained_names, self._name
        )
        name, status, start, stop, pretrain_layout = (
            pretrain_layout_group.name,
            pretrain_layout_group.status,
            pretrain_layout_group.start,
            pretrain_layout_group.stop,
            pretrain_layout_group.layout,
        )

        lora_layout_group = create_lora_layout(self.supported_lora_names, num_loras=5)
        loras = lora_layout_group.loras
        lora_layout = lora_layout_group.layout
        lora0, lora1, lora2, lora3, lora4 = loras
        (
            lora0_checkpoint,
            lora0_weight,
            lora0_alpha,
            lora0_url,
            lora0_file,
        ) = (
            lora0.checkpoint,
            lora0.weight,
            lora0.alpha,
            lora0.url,
            lora0.file,
        )
        (
            lora1_checkpoint,
            lora1_weight,
            lora1_alpha,
            lora1_url,
            lora1_file,
        ) = (
            lora1.checkpoint,
            lora1.weight,
            lora1.alpha,
            lora1.url,
            lora1.file,
        )
        (
            lora2_checkpoint,
            lora2_weight,
            lora2_alpha,
            lora2_url,
            lora2_file,
        ) = (
            lora2.checkpoint,
            lora2.weight,
            lora2.alpha,
            lora2.url,
            lora2.file,
        )
        (
            lora3_checkpoint,
            lora3_weight,
            lora3_alpha,
            lora3_url,
            lora3_file,
        ) = (
            lora3.checkpoint,
            lora3.weight,
            lora3.alpha,
            lora3.url,
            lora3.file,
        )
        (
            lora4_checkpoint,
            lora4_weight,
            lora4_alpha,
            lora4_url,
            lora4_file,
        ) = (
            lora4.checkpoint,
            lora4.weight,
            lora4.alpha,
            lora4.url,
            lora4.file,
        )

        prompt = create_element("text", "Input Prompt", lines=3)
        generate = create_element("button", "Generate")
        result = create_element("text", "Output Result", lines=3)

        # Create the blocks
        left = create_column(prompt, lora_layout, generate)
        right = create_column(result)
        iface = create_blocks(pretrain_layout, create_row(left, right))

        # Create the events
        iface.__enter__()

        start.click(self.start, inputs=[name], outputs=[status])
        stop.click(self.stop, outputs=[status])
        generate.click(
            self.serve,
            inputs=[
                prompt,
                lora0_checkpoint,
                lora0_weight,
                lora0_alpha,
                lora0_url,
                lora0_file,
                lora1_checkpoint,
                lora1_weight,
                lora1_alpha,
                lora1_url,
                lora1_file,
                lora2_checkpoint,
                lora2_weight,
                lora2_alpha,
                lora2_url,
                lora2_file,
                lora3_checkpoint,
                lora3_weight,
                lora3_alpha,
                lora3_url,
                lora3_file,
                lora4_checkpoint,
                lora4_weight,
                lora4_alpha,
                lora4_url,
                lora4_file,
            ],
            outputs=[result],
        )

        iface.load(
            fn=lambda: [gr.update(value=self._name), gr.update(value=self._status)],
            outputs=[name, status],
        )

        iface.__exit__()

        super().__init__(config, iname="LLAMA", iface=iface)

    def start(self, pretrained_name, **kwargs):
        if self._status == "Running":
            self.stop()
        self._name = pretrained_name
        self._pipe = LlamaForGenerationPipeline.from_core_configure(
            self._config,
            pretrained_name=pretrained_name,
        )
        self._status = "Running"
        return self._status

    def stop(self, **kwargs):
        self._pipe.to("cpu")
        del self._pipe
        gc.collect()
        torch.cuda.empty_cache()
        self._pipe = None if not hasattr(self, "_pipe") else self._pipe
        self._status = "Stopped" if self._pipe is None else "Running"
        return self._status

    def serve(
        self,
        text: str,
        lora0_checkpoint: Optional[str] = None,
        lora0_weight: Optional[float] = 1.0,
        lora0_alpha: Optional[float] = 32,
        lora0_url: Optional[str] = None,
        lora0_file: Optional[str] = None,
        lora1_checkpoint: Optional[str] = None,
        lora1_weight: Optional[float] = 1.0,
        lora1_alpha: Optional[float] = 32,
        lora1_url: Optional[str] = None,
        lora1_file: Optional[str] = None,
        lora2_checkpoint: Optional[str] = None,
        lora2_weight: Optional[float] = 1.0,
        lora2_alpha: Optional[float] = 32,
        lora2_url: Optional[str] = None,
        lora2_file: Optional[str] = None,
        lora3_checkpoint: Optional[str] = None,
        lora3_weight: Optional[float] = 1.0,
        lora3_alpha: Optional[float] = 32,
        lora3_url: Optional[str] = None,
        lora3_file: Optional[str] = None,
        lora4_checkpoint: Optional[str] = None,
        lora4_weight: Optional[float] = 1.0,
        lora4_alpha: Optional[float] = 32,
        lora4_url: Optional[str] = None,
        lora4_file: Optional[str] = None,
    ):
        assert self._pipe is not None
        result = self._pipe(
            text,
            lora_checkpoints=(
                lora0_checkpoint,
                lora1_checkpoint,
                lora2_checkpoint,
                lora3_checkpoint,
                lora4_checkpoint,
            ),
            lora_weights=(
                lora0_weight,
                lora1_weight,
                lora2_weight,
                lora3_weight,
                lora4_weight,
            ),
            lora_alphas=(
                lora0_alpha,
                lora1_alpha,
                lora2_alpha,
                lora3_alpha,
                lora4_alpha,
            ),
            lora_urls=(
                lora0_url,
                lora1_url,
                lora2_url,
                lora3_url,
                lora4_url,
            ),
            lora_files=(
                lora0_file,
                lora1_file,
                lora2_file,
                lora3_file,
                lora4_file,
            ),
        )
        return result



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import re
import torch
import numpy as np
from PIL import Image, ImageDraw
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from unitorch.models.detr import (
    DetrForDetection as _DetrForDetection,
)
from unitorch.models.detr import DetrProcessor
from unitorch.utils import pop_value, nested_dict_value
from unitorch.cli import (
    cached_path,
    add_default_section_for_init,
    add_default_section_for_function,
)
from unitorch.cli.models.detr import pretrained_detr_infos


class DetrForDetectionPipeline(_DetrForDetection):
    def __init__(
        self,
        config_path: str,
        vision_config_path: str,
        weight_path: Optional[Union[str, List[str]]] = None,
        state_dict: Optional[Dict[str, Any]] = None,
        device: Optional[Union[str, int]] = "cpu",
    ):
        super().__init__(
            config_path=config_path,
        )
        self.processor = DetrProcessor(
            vision_config_path=vision_config_path,
        )
        self._device = "cpu" if device == "cpu" else int(device)

        self.from_pretrained(weight_path, state_dict=state_dict)
        self.to(device=self._device)
        self.eval()

    @classmethod
    @add_default_section_for_init("core/pipeline/detr")
    def from_core_configure(
        cls,
        config,
        pretrained_name: Optional[str] = "default-detr",
        config_path: Optional[str] = None,
        vision_config_path: Optional[str] = None,
        pretrained_weight_path: Optional[str] = None,
        device: Optional[str] = "cpu",
        **kwargs,
    ):
        config.set_default_section("core/pipeline/detr")
        pretrained_name = config.getoption("pretrained_name", pretrained_name)

        config_path = config.getoption("config_path", config_path)
        config_path = pop_value(
            config_path,
            nested_dict_value(pretrained_detr_infos, pretrained_name, "config"),
        )
        config_path = cached_path(config_path)

        vision_config_path = config.getoption("vision_config_path", vision_config_path)
        vision_config_path = pop_value(
            vision_config_path,
            nested_dict_value(pretrained_detr_infos, pretrained_name, "vision_config"),
        )
        vision_config_path = cached_path(vision_config_path)

        device = config.getoption("device", device)
        pretrained_weight_path = config.getoption(
            "pretrained_weight_path", pretrained_weight_path
        )
        weight_path = pop_value(
            pretrained_weight_path,
            nested_dict_value(pretrained_detr_infos, pretrained_name, "weight"),
            check_none=False,
        )

        inst = cls(
            config_path,
            vision_config_path,
            weight_path=weight_path,
            device=device,
        )

        return inst

    @torch.no_grad()
    @add_default_section_for_function("core/pipeline/detr")
    def __call__(
        self,
        image: Union[Image.Image, str],
        threshold: Optional[float] = 0.5,
    ):
        inputs = self.processor.image(image)
        pixel_values = [inputs.image.to(self._device)]
        outputs = self.detect(
            pixel_values,
            norm_bboxes=True,
            threshold=threshold,
        )
        result_image = image.copy()
        bboxes = outputs["bboxes"][0].cpu().numpy()
        scores = outputs["scores"][0].cpu().numpy()
        classes = outputs["classes"][0].cpu().numpy()
        # draw boxes on the result image with class and score
        for bbox, score, classid in zip(bboxes, scores, classes):
            if score < threshold:
                continue
            bbox = bbox * np.array(
                [
                    result_image.width,
                    result_image.height,
                    result_image.width,
                    result_image.height,
                ]
            )
            bbox = list(map(int, bbox))
            result_image = result_image.copy()
            draw = ImageDraw.Draw(result_image)
            draw.rectangle(bbox, outline="red", width=3)
            draw.text((bbox[0] + 5, bbox[1] + 5), f"{classid}", fill="blue")
            draw.text((bbox[0] + 5, bbox[1] + 15), f"{score:.2f}", fill="green")

        return result_image



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import gradio as gr
from unitorch.cli.webuis.utils.layouts import (
    create_element,
    create_accordion,
    create_row,
    create_column,
    create_group,
    create_tab,
    create_tabs,
    create_blocks,
    create_pretrain_layout,
)
from unitorch.cli import register_webui
from unitorch.cli import CoreConfigureParser, GenericWebUI
from unitorch.cli.webuis import SimpleWebUI
from unitorch.cli.webuis.tools.image import ImageWebUI


@register_webui("core/webui/tools")
class ToolsWebUI(SimpleWebUI):
    def __init__(self, config: CoreConfigureParser):
        webuis = [
            ImageWebUI(config),
        ]
        iface = gr.TabbedInterface(
            [webui.iface for webui in webuis],
            tab_names=[webui.iname for webui in webuis],
        )
        super().__init__(config, iname="Tools", iface=iface)



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import re
import torch
from PIL import Image
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from unitorch.models import GenericOutputs
from unitorch.models.clip import (
    ClipForPretrain as _ClipForPretrain,
    ClipProcessor,
)
from unitorch.utils import pop_value, nested_dict_value, read_file, read_json_file

from unitorch.cli import (
    get_global_config,
    cached_path,
    add_default_section_for_init,
    add_default_section_for_function,
)
from unitorch.cli.models.clip import pretrained_clip_infos
from unitorch.cli.pipelines.blip import BlipForImageCaptionPipeline


class ClipInterrogatorPipeline(_ClipForPretrain):
    def __init__(
        self,
        config_path: str,
        vocab_path: str,
        merge_path: str,
        vision_config_path: str,
        max_seq_length: Optional[int] = 77,
        weight_path: Optional[Union[str, List[str]]] = None,
        state_dict: Optional[Dict[str, Any]] = None,
        device: Optional[Union[str, int]] = "cpu",
    ):
        projection_dim = nested_dict_value(
            read_json_file(config_path), "projection_dim"
        )
        super().__init__(
            config_path=config_path,
            projection_dim=projection_dim,
        )
        self.processor = ClipProcessor(
            vocab_path=vocab_path,
            merge_path=merge_path,
            vision_config_path=vision_config_path,
            max_seq_length=max_seq_length,
        )
        self._device = "cpu" if device == "cpu" else int(device)

        self.from_pretrained(weight_path, state_dict=state_dict)
        self.to(device=self._device)

        config = get_global_config()
        self.blip_pipe = BlipForImageCaptionPipeline.from_core_configure(
            config, pretrained_name="blip-image-captioning-large"
        )
        self.blip_pipe._device = self._device
        self.blip_pipe.to(device=self._device)

        artists = read_file(
            cached_path(
                "https://huggingface.co/datasets/fuliucansheng/hubfiles/raw/main/clip-interrogator/artists.txt"
            ),
            lines=True,
        )
        flavors = read_file(
            cached_path(
                "https://huggingface.co/datasets/fuliucansheng/hubfiles/raw/main/clip-interrogator/flavors.txt"
            ),
            lines=True,
        )
        mediums = read_file(
            cached_path(
                "https://huggingface.co/datasets/fuliucansheng/hubfiles/raw/main/clip-interrogator/mediums.txt"
            ),
            lines=True,
        )
        movements = read_file(
            cached_path(
                "https://huggingface.co/datasets/fuliucansheng/hubfiles/raw/main/clip-interrogator/movements.txt"
            ),
            lines=True,
        )
        negative = read_file(
            cached_path(
                "https://huggingface.co/datasets/fuliucansheng/hubfiles/raw/main/clip-interrogator/negative.txt"
            ),
            lines=True,
        )
        sites = read_file(
            cached_path(
                "https://huggingface.co/datasets/fuliucansheng/hubfiles/raw/main/clip-interrogator/sites.txt"
            ),
            lines=True,
        )

        artists = [f"by {a}" for a in artists] + [f"inspired by {a}" for a in artists]
        trendings = (
            sites
            + [f"trending on {s}" for s in sites]
            + [f"featured on {s}" for s in sites]
            + [f"{s} contest winner" for s in sites]
        )

        self.positive_labels = artists + flavors + mediums + movements + trendings
        self.positive_artists_labels = artists
        self.positive_flavors_labels = flavors
        self.positive_mediums_labels = mediums
        self.positive_movements_labels = movements
        self.positive_trendings_labels = trendings
        self.negative_labels = negative

        self.eval()

        self.positive_labels_embeds = self.get_text_embeds(self.positive_labels)

    @classmethod
    @add_default_section_for_init("core/pipeline/interrogator/clip")
    def from_core_configure(
        cls,
        config,
        pretrained_name: Optional[str] = "default-clip",
        config_path: Optional[str] = None,
        vocab_path: Optional[str] = None,
        merge_path: Optional[str] = None,
        vision_config_path: Optional[str] = None,
        pretrained_weight_path: Optional[str] = None,
        device: Optional[str] = "cpu",
        **kwargs,
    ):
        config.set_default_section("core/pipeline/interrogator/clip")
        pretrained_name = config.getoption("pretrained_name", pretrained_name)

        config_path = config.getoption("config_path", config_path)
        config_path = pop_value(
            config_path,
            nested_dict_value(pretrained_clip_infos, pretrained_name, "config"),
        )
        config_path = cached_path(config_path)

        vocab_path = config.getoption("vocab_path", vocab_path)
        vocab_path = pop_value(
            vocab_path,
            nested_dict_value(pretrained_clip_infos, pretrained_name, "vocab"),
        )
        vocab_path = cached_path(vocab_path)

        merge_path = config.getoption("merge_path", merge_path)
        merge_path = pop_value(
            merge_path,
            nested_dict_value(pretrained_clip_infos, pretrained_name, "merge"),
        )
        merge_path = cached_path(merge_path)

        vision_config_path = config.getoption("vision_config_path", vision_config_path)
        vision_config_path = pop_value(
            vision_config_path,
            nested_dict_value(pretrained_clip_infos, pretrained_name, "vision_config"),
        )
        vision_config_path = cached_path(vision_config_path)

        max_seq_length = config.getoption("max_seq_length", 77)
        device = config.getoption("device", device)
        pretrained_weight_path = config.getoption(
            "pretrained_weight_path", pretrained_weight_path
        )
        weight_path = pop_value(
            pretrained_weight_path,
            nested_dict_value(pretrained_clip_infos, pretrained_name, "weight"),
            check_none=False,
        )

        inst = cls(
            config_path,
            vocab_path,
            merge_path,
            vision_config_path,
            max_seq_length=max_seq_length,
            weight_path=weight_path,
            device=device,
        )

        return inst

    def get_image_embeds(self, image: Image.Image):
        inputs = self.processor.image_classification(image)
        inputs = {k: v.unsqueeze(0) if v is not None else v for k, v in inputs.items()}
        inputs = {
            k: v.to(device=self._device) if v is not None else v
            for k, v in inputs.items()
        }
        vision_outputs = self.vision_model(pixel_values=inputs["pixel_values"])
        image_embeds = self.visual_projection(vision_outputs[1])
        image_embeds = image_embeds / image_embeds.norm(dim=-1, keepdim=True)
        return image_embeds.cpu()

    def get_text_embeds(self, texts: Union[str, List[str]], max_batch_size=1024):
        if isinstance(texts, str):
            texts = [texts]
        inputs = [self.processor.text_classification(text) for text in texts]
        keys = inputs[0].keys()
        inputs = {
            k: torch.stack([i[k] for i in inputs]).to(device=self._device) for k in keys
        }
        results = []
        for i in range(0, len(inputs["input_ids"]), max_batch_size):
            text_outputs = self.text_model(
                input_ids=inputs["input_ids"][i : i + max_batch_size],
                attention_mask=inputs["attention_mask"][i : i + max_batch_size],
                position_ids=inputs["position_ids"][i : i + max_batch_size],
            )
            text_embeds = self.text_projection(text_outputs[1])
            text_embeds = text_embeds / text_embeds.norm(dim=-1, keepdim=True)
            results.append(text_embeds.cpu())
        return torch.cat(results, dim=0)

    def get_score(self, image_embeds, text):
        text_embeds = self.get_text_embeds(text)
        scores = torch.einsum("te,be->tb", text_embeds, image_embeds)
        return scores[0][0].item()

    def rank_top(self, image_embeds, texts=[], topk=1, reverse=True, text_embeds=None):
        if text_embeds is None:
            text_embeds = self.get_text_embeds(texts)
        scores = torch.einsum("te,be->tb", text_embeds, image_embeds)
        scores = scores[:, 0].tolist()
        return sorted(list(zip(texts, scores)), key=lambda x: x[1], reverse=reverse)[
            :topk
        ]

    def chain(
        self,
        image_embeds,
        phrases,
        caption=None,
        min_count=8,
        max_count=32,
        reverse=True,
    ):
        if caption is None:
            caption = self.rank_top(image_embeds, phrases, topk=1, reverse=reverse)[0][
                0
            ]
        best_prompt, best_score = caption, self.get_score(image_embeds, caption)
        if best_prompt in phrases:
            phrases.remove(best_prompt)

        prompt, score = best_prompt, best_score

        for i in range(max_count):
            new_prompt, new_score = self.rank_top(
                image_embeds,
                [f"{prompt}, {p}" for p in phrases],
                topk=1,
                reverse=reverse,
            )[0]

            label = new_prompt[len(prompt) + 2 :]
            phrases.remove(label)

            if (reverse and new_score > best_score) or (
                not reverse and new_score < best_score
            ):
                best_prompt, best_score = new_prompt, new_score
            if (
                i < min_count
                or (reverse and new_score < score)
                or (not reverse and new_score > score)
            ):
                prompt, score = new_prompt, new_score
            else:
                break

        return best_prompt

    def get_best_prompt(self, image_embeds, caption="", min_count=8, max_count=32):
        positive_labels = self.rank_top(
            image_embeds,
            texts=self.positive_labels,
            topk=1024,
            text_embeds=self.positive_labels_embeds,
        )
        positive_labels = list(map(lambda x: x[0], positive_labels))

        best_prompt = self.chain(
            image_embeds, positive_labels, caption, min_count, max_count
        )
        return best_prompt

    def get_fast_prompt(self, image_embeds, caption="", max_count=32):
        positive_labels = self.rank_top(
            image_embeds,
            texts=self.positive_labels,
            topk=max_count,
            text_embeds=self.positive_labels_embeds,
        )
        positive_labels = list(map(lambda x: x[0], positive_labels))

        fast_prompt = f"{caption}, {', '.join(positive_labels)}"
        return fast_prompt

    def get_classic_prompt(self, image_embeds, caption="", max_count=3):
        positive_embeds = self.positive_labels_embeds.clone()
        positive_artists_labels_embeds, positive_embeds = (
            positive_embeds[: len(self.positive_artists_labels)],
            positive_embeds[len(self.positive_artists_labels) :],
        )
        positive_flavors_labels_embeds, positive_embeds = (
            positive_embeds[: len(self.positive_flavors_labels)],
            positive_embeds[len(self.positive_flavors_labels) :],
        )
        positive_mediums_labels_embeds, positive_embeds = (
            positive_embeds[: len(self.positive_mediums_labels)],
            positive_embeds[len(self.positive_mediums_labels) :],
        )
        positive_movements_labels_embeds, positive_embeds = (
            positive_embeds[: len(self.positive_movements_labels)],
            positive_embeds[len(self.positive_movements_labels) :],
        )
        positive_trendings_labels_embeds, positive_embeds = (
            positive_embeds[: len(self.positive_trendings_labels)],
            positive_embeds[len(self.positive_trendings_labels) :],
        )

        artist = self.rank_top(
            image_embeds,
            texts=self.positive_artists_labels,
            topk=1,
            text_embeds=positive_artists_labels_embeds,
        )[0][0]
        flavors = self.rank_top(
            image_embeds,
            texts=self.positive_flavors_labels,
            topk=max_count,
            text_embeds=positive_flavors_labels_embeds,
        )
        flavors = ", ".join(list(map(lambda x: x[0], flavors)))
        medium = self.rank_top(
            image_embeds,
            texts=self.positive_mediums_labels,
            topk=1,
            text_embeds=positive_mediums_labels_embeds,
        )[0][0]
        movement = self.rank_top(
            image_embeds,
            texts=self.positive_movements_labels,
            topk=1,
            text_embeds=positive_movements_labels_embeds,
        )[0][0]
        trending = self.rank_top(
            image_embeds,
            texts=self.positive_trendings_labels,
            topk=1,
            text_embeds=positive_trendings_labels_embeds,
        )[0][0]

        if caption.startswith(medium):
            classic_prompt = f"{caption} {artist}, {trending}, {movement}, {flavors}"
        else:
            classic_prompt = (
                f"{caption}, {medium} {artist}, {trending}, {movement}, {flavors}"
            )

        return classic_prompt

    def get_negative_prompt(self, image_embeds, max_count=3):
        positive_flavors_labels_embeds = self.positive_labels_embeds[
            len(self.positive_artists_labels) : len(self.positive_artists_labels)
            + len(self.positive_flavors_labels)
        ]
        negative_labels = self.rank_top(
            image_embeds,
            texts=self.positive_flavors_labels,
            topk=max_count,
            reverse=False,
            text_embeds=positive_flavors_labels_embeds,
        )
        negative_labels = list(map(lambda x: x[0], negative_labels))
        negative_labels += self.negative_labels
        negative_prompt = self.chain(
            image_embeds, negative_labels, max_count=max_count, reverse=False
        )
        return negative_prompt

    @torch.no_grad()
    @add_default_section_for_function("core/pipeline/interrogator/clip")
    def __call__(
        self,
        image: Image.Image,
    ):
        caption = self.blip_pipe(image, num_beams=5, max_gen_seq_length=32)
        image_embeds = self.get_image_embeds(image)
        fast_prompt = self.get_fast_prompt(image_embeds, caption)
        classic_prompt = self.get_classic_prompt(image_embeds, caption)
        best_prompt = self.get_best_prompt(image_embeds, caption)
        negative_prompt = self.get_negative_prompt(image_embeds)
        return GenericOutputs(
            fast_prompt=fast_prompt,
            classic_prompt=classic_prompt,
            best_prompt=best_prompt,
            negative_prompt=negative_prompt,
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import torch
import logging
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from unitorch.optims import SGD, Adam, AdamW
from unitorch.models import CheckpointMixin
from unitorch.cli import add_default_section_for_init, register_optim


@register_optim("core/optim/sgd")
class SGDOptimizer(SGD, CheckpointMixin):
    def __init__(
        self,
        params,
        learning_rate: Optional[float] = 0.00001,
    ):
        super().__init__(
            params=params,
            lr=learning_rate,
        )

    @classmethod
    @add_default_section_for_init("core/optim/sgd")
    def from_core_configure(cls, config, **kwargs):
        pass


@register_optim("core/optim/adam")
class AdamOptimizer(Adam, CheckpointMixin):
    def __init__(
        self,
        params,
        learning_rate: Optional[float] = 0.00001,
    ):
        super().__init__(
            params=params,
            lr=learning_rate,
        )

    @classmethod
    @add_default_section_for_init("core/optim/adam")
    def from_core_configure(cls, config, **kwargs):
        pass


@register_optim("core/optim/adamw")
class AdamWOptimizer(AdamW, CheckpointMixin):
    def __init__(
        self,
        params,
        learning_rate: Optional[float] = 0.00001,
    ):
        super().__init__(
            params=params,
            lr=learning_rate,
        )

    @classmethod
    @add_default_section_for_init("core/optim/adamw")
    def from_core_configure(cls, config, **kwargs):
        pass


# more optims
import unitorch.cli.optims.lion

# bitsandbytes
from unitorch.utils import is_bitsandbytes_available

if is_bitsandbytes_available():
    from unitorch.optims import Adam8bit, AdamW8bit

    @register_optim("core/optim/adam8bit")
    class Adam8bitOptimizer(Adam8bit, CheckpointMixin):
        def __init__(
            self,
            params,
            learning_rate: Optional[float] = 0.00001,
        ):
            super().__init__(
                params=params,
                lr=learning_rate,
            )

        @classmethod
        @add_default_section_for_init("core/optim/adam8bit")
        def from_core_configure(cls, config, **kwargs):
            pass

    @register_optim("core/optim/adamw8bit")
    class AdamW8bitOptimizer(AdamW8bit, CheckpointMixin):
        def __init__(
            self,
            params,
            learning_rate: Optional[float] = 0.00001,
        ):
            super().__init__(
                params=params,
                lr=learning_rate,
            )

        @classmethod
        @add_default_section_for_init("core/optim/adamw8bit")
        def from_core_configure(cls, config, **kwargs):
            pass



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import io
import torch
import gc
import gradio as gr
from PIL import Image
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from unitorch.cli import CoreConfigureParser, GenericWebUI
from unitorch.cli import register_webui
from unitorch.cli.models.diffusers import (
    pretrained_stable_infos,
    pretrained_stable_extensions_infos,
)
from unitorch.cli.pipelines.stable import StableForImage2VideoGenerationPipeline
from unitorch.cli.pipelines.stable import controlnet_processes
from unitorch.cli.webuis import (
    supported_scheduler_names,
    matched_pretrained_names,
    create_element,
    create_accordion,
    create_row,
    create_column,
    create_group,
    create_tab,
    create_tabs,
    create_blocks,
    create_pretrain_layout,
    create_controlnet_layout,
    create_lora_layout,
    create_freeu_layout,
)
from unitorch.cli.webuis import SimpleWebUI


class StableImage2VideoWebUI(SimpleWebUI):
    pretrained_names = list(pretrained_stable_infos.keys())
    supported_pretrained_names = matched_pretrained_names(
        pretrained_names, "^stable-video-"
    )
    pretrained_extension_names = list(pretrained_stable_extensions_infos.keys())
    supported_lora_names = matched_pretrained_names(
        pretrained_extension_names,
        "^stable-video-lora-",
    )
    supported_schedulers = supported_scheduler_names

    def __init__(self, config: CoreConfigureParser):
        self._pipe = None if not hasattr(self, "_pipe") else self._pipe
        self._status = "Stopped" if self._pipe is None else "Running"
        if len(self.supported_pretrained_names) == 0:
            raise ValueError("No supported pretrained models found.")
        self._name = self.supported_pretrained_names[0]

        # create elements
        pretrain_layout_group = create_pretrain_layout(
            self.supported_pretrained_names, self._name
        )
        name, status, start, stop, pretrain_layout = (
            pretrain_layout_group.name,
            pretrain_layout_group.status,
            pretrain_layout_group.start,
            pretrain_layout_group.stop,
            pretrain_layout_group.layout,
        )

        scheduler = create_element(
            "dropdown",
            "Sampling Method",
            values=self.supported_schedulers,
            default=self.supported_schedulers[0],
        )
        steps = create_element(
            "slider", "Diffusion Steps", min_value=1, max_value=100, step=1, default=25
        )
        image = create_element("image", "Input Image")
        height = create_element(
            "slider", "Video Height", min_value=1, max_value=2048, step=1, default=576
        )
        width = create_element(
            "slider", "Video Width", min_value=1, max_value=2048, step=1, default=1024
        )

        num_frames = create_element(
            "slider", "Video Frames", min_value=1, max_value=100, step=1, default=30
        )
        num_fps = gr.Slider(1, 100, value=7, label="Video FPS", step=1)

        min_guidance_scale = create_element(
            "slider",
            "Min Guidance Scale",
            min_value=0,
            max_value=10,
            step=0.1,
            default=1.0,
        )
        max_guidance_scale = create_element(
            "slider",
            "Max Guidance Scale",
            min_value=0,
            max_value=10,
            step=0.1,
            default=2.5,
        )
        motion_bucket_id = gr.Slider(
            1, 255, value=127, label="Video Motion Bucket Id", step=1
        )
        decode_chunk_size = gr.Slider(1, 32, value=8, label="Decode Chunk Size", step=1)
        seed = create_element(
            "slider", "Seed", min_value=0, max_value=9999, step=1, default=42
        )

        ## extensions
        lora_layout_group = create_lora_layout(self.supported_lora_names, num_loras=5)
        loras = lora_layout_group.loras
        lora_layout = lora_layout_group.layout
        lora0, lora1, lora2, lora3, lora4 = loras
        (
            lora0_checkpoint,
            lora0_weight,
            lora0_alpha,
            lora0_url,
            lora0_file,
        ) = (
            lora0.checkpoint,
            lora0.weight,
            lora0.alpha,
            lora0.url,
            lora0.file,
        )
        (
            lora1_checkpoint,
            lora1_weight,
            lora1_alpha,
            lora1_url,
            lora1_file,
        ) = (
            lora1.checkpoint,
            lora1.weight,
            lora1.alpha,
            lora1.url,
            lora1.file,
        )
        (
            lora2_checkpoint,
            lora2_weight,
            lora2_alpha,
            lora2_url,
            lora2_file,
        ) = (
            lora2.checkpoint,
            lora2.weight,
            lora2.alpha,
            lora2.url,
            lora2.file,
        )
        (
            lora3_checkpoint,
            lora3_weight,
            lora3_alpha,
            lora3_url,
            lora3_file,
        ) = (
            lora3.checkpoint,
            lora3.weight,
            lora3.alpha,
            lora3.url,
            lora3.file,
        )
        (
            lora4_checkpoint,
            lora4_weight,
            lora4_alpha,
            lora4_url,
            lora4_file,
        ) = (
            lora4.checkpoint,
            lora4.weight,
            lora4.alpha,
            lora4.url,
            lora4.file,
        )

        generate = create_element("button", "Generate", variant="primary", scale=2)
        output_video = create_element("video", "Output Video")

        # create layouts
        top = create_column(pretrain_layout)
        left_generation = create_tab(
            create_row(image),
            create_row(height),
            create_row(width),
            create_row(num_frames, num_fps),
            create_row(min_guidance_scale),
            create_row(max_guidance_scale),
            create_row(motion_bucket_id),
            create_row(decode_chunk_size),
            create_row(scheduler, steps),
            create_row(seed),
            create_row(generate),
            name="Generation",
        )
        left_extension = create_tab(
            create_row(lora_layout),
            name="Extensions",
        )
        left_settings = create_tab(
            name="Settings",
        )

        left = create_tabs(left_generation, left_extension, left_settings)
        right = create_column(output_video)
        iface = create_blocks(top, create_row(left, right))

        # create events
        iface.__enter__()

        start.click(fn=self.start, inputs=[name], outputs=[status])
        stop.click(fn=self.stop, outputs=[status])

        generate.click(
            fn=self.serve,
            inputs=[
                image,
                height,
                width,
                num_frames,
                num_fps,
                min_guidance_scale,
                max_guidance_scale,
                motion_bucket_id,
                decode_chunk_size,
                steps,
                seed,
                scheduler,
                lora0_checkpoint,
                lora0_weight,
                lora0_alpha,
                lora0_url,
                lora0_file,
                lora1_checkpoint,
                lora1_weight,
                lora1_alpha,
                lora1_url,
                lora1_file,
                lora2_checkpoint,
                lora2_weight,
                lora2_alpha,
                lora2_url,
                lora2_file,
                lora3_checkpoint,
                lora3_weight,
                lora3_alpha,
                lora3_url,
                lora3_file,
                lora4_checkpoint,
                lora4_weight,
                lora4_alpha,
                lora4_url,
                lora4_file,
            ],
            outputs=[output_video],
        )
        iface.load(
            fn=lambda: [gr.update(value=self._name), gr.update(value=self._status)],
            outputs=[name, status],
        )

        iface.__exit__()

        super().__init__(config, iname="Image2Video", iface=iface)

    def start(self, pretrained_name, **kwargs):
        if self._status == "Running":
            self.stop()
        self._name = pretrained_name
        self._pipe = StableForImage2VideoGenerationPipeline.from_core_configure(
            self._config,
            pretrained_name=pretrained_name,
        )
        self._status = "Running"
        return self._status

    def stop(self, **kwargs):
        self._pipe.to("cpu")
        del self._pipe
        gc.collect()
        torch.cuda.empty_cache()
        self._pipe = None if not hasattr(self, "_pipe") else self._pipe
        self._status = "Stopped" if self._pipe is None else "Running"
        return self._status

    def serve(
        self,
        image: Image.Image,
        height: Optional[int] = 576,
        width: Optional[int] = 1024,
        num_frames: Optional[int] = 30,
        num_fps: Optional[int] = 6,
        min_guidance_scale: Optional[float] = 1.0,
        max_guidance_scale: Optional[float] = 2.5,
        motion_bucket_id: Optional[int] = 127,
        decode_chunk_size: Optional[int] = 8,
        num_timesteps: Optional[int] = 25,
        seed: Optional[int] = 1123,
        scheduler: Optional[str] = None,
        lora0_checkpoint: Optional[str] = None,
        lora0_weight: Optional[float] = 1.0,
        lora0_alpha: Optional[float] = 32,
        lora0_url: Optional[str] = None,
        lora0_file: Optional[str] = None,
        lora1_checkpoint: Optional[str] = None,
        lora1_weight: Optional[float] = 1.0,
        lora1_alpha: Optional[float] = 32,
        lora1_url: Optional[str] = None,
        lora1_file: Optional[str] = None,
        lora2_checkpoint: Optional[str] = None,
        lora2_weight: Optional[float] = 1.0,
        lora2_alpha: Optional[float] = 32,
        lora2_url: Optional[str] = None,
        lora2_file: Optional[str] = None,
        lora3_checkpoint: Optional[str] = None,
        lora3_weight: Optional[float] = 1.0,
        lora3_alpha: Optional[float] = 32,
        lora3_url: Optional[str] = None,
        lora3_file: Optional[str] = None,
        lora4_checkpoint: Optional[str] = None,
        lora4_weight: Optional[float] = 1.0,
        lora4_alpha: Optional[float] = 32,
        lora4_url: Optional[str] = None,
        lora4_file: Optional[str] = None,
    ):
        assert self._pipe is not None
        video = self._pipe(
            image,
            height=height,
            width=width,
            num_frames=num_frames,
            num_fps=num_fps,
            min_guidance_scale=min_guidance_scale,
            max_guidance_scale=max_guidance_scale,
            motion_bucket_id=motion_bucket_id,
            decode_chunk_size=decode_chunk_size,
            num_timesteps=num_timesteps,
            seed=seed,
            scheduler=scheduler,
            lora_checkpoints=(
                lora0_checkpoint,
                lora1_checkpoint,
                lora2_checkpoint,
                lora3_checkpoint,
                lora4_checkpoint,
            ),
            lora_weights=(
                lora0_weight,
                lora1_weight,
                lora2_weight,
                lora3_weight,
                lora4_weight,
            ),
            lora_alphas=(
                lora0_alpha,
                lora1_alpha,
                lora2_alpha,
                lora3_alpha,
                lora4_alpha,
            ),
            lora_urls=(
                lora0_url,
                lora1_url,
                lora2_url,
                lora3_url,
                lora4_url,
            ),
            lora_files=(
                lora0_file,
                lora1_file,
                lora2_file,
                lora3_file,
                lora4_file,
            ),
        )
        return video



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

# pretrained infos

pretrained_visualbert_infos = {
    "default-visualbert": {
        "config": "https://huggingface.co/uclanlp/visualbert-vqa-coco-pre/resolve/main/config.json",
        "vocab": "https://huggingface.co/bert-base-uncased/resolve/main/vocab.txt",
    },
    "visualbert-vqa-coco-pre": {
        "config": "https://huggingface.co/uclanlp/visualbert-vqa-coco-pre/resolve/main/config.json",
        "vocab": "https://huggingface.co/bert-base-uncased/resolve/main/vocab.txt",
        "weight": "https://huggingface.co/uclanlp/visualbert-vqa-coco-pre/resolve/main/pytorch_model.bin",
    },
    "visualbert-nlvr2-coco-pre": {
        "config": "https://huggingface.co/uclanlp/visualbert-nlvr2-coco-pre/resolve/main/config.json",
        "vocab": "https://huggingface.co/bert-base-multilingual-uncased/resolve/main/vocab.txt",
        "weight": "https://huggingface.co/uclanlp/visualbert-nlvr2-coco-pre/resolve/main/pytorch_model.bin",
    },
}

import unitorch.cli.models.visualbert.modeling
import unitorch.cli.models.visualbert.processing
from unitorch.cli.models.visualbert.modeling import (
    VisualBertForClassification,
    VisualBertForPretrain,
)
from unitorch.cli.models.visualbert.processing import VisualBertProcessor



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import re
import json
import torch
import logging
import torch.nn as nn
import safetensors
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from collections import OrderedDict
from transformers.utils import is_remote_url, ModelOutput as GenericOutputs
from unitorch import hf_cached_path
from unitorch.utils import replace, load_weight, is_diffusers_available


class CheckpointMixin:
    checkpoint_name = "pytorch_model.bin"
    replace_keys_in_state_dict = {}
    prefix_keys_in_state_dict = {}

    def from_checkpoint(
        self,
        ckpt_dir: str,
        weight_name: Optional[str] = None,
        **kwargs,
    ):
        """
        Load model weights from a checkpoint.

        Args:
            ckpt_dir (str): Directory path of the checkpoint.
            weight_name (str): Name of the weight file.

        Returns:
            None
        """
        if weight_name is None:
            weight_name = self.checkpoint_name
        weight_path = os.path.join(ckpt_dir, weight_name)
        if not os.path.exists(weight_path):
            return

        if weight_path.endswith(".safetensors"):
            state_dict = safetensors.torch.load_file(weight_path)
        else:
            state_dict = torch.load(weight_path, map_location="cpu")
        self.load_state_dict(state_dict)
        logging.info(
            f"{type(self).__name__} model load weight from checkpoint {weight_path}"
        )

    def save_checkpoint(
        self,
        ckpt_dir: str,
        weight_name: Optional[str] = None,
        **kwargs,
    ):
        """
        Save the model's current state as a checkpoint.

        Args:
            ckpt_dir (str): Directory path to save the checkpoint.
            weight_name (str): Name of the weight file.

        Returns:
            None
        """
        if weight_name is None:
            weight_name = self.checkpoint_name
        state_dict = self.state_dict()
        weight_path = os.path.join(ckpt_dir, weight_name)
        if weight_path.endswith(".safetensors"):
            safetensors.torch.save_file(state_dict, weight_path)
        else:
            torch.save(state_dict, weight_path)
        logging.info(f"{type(self).__name__} model save checkpoint to {weight_path}")

    def from_pretrained(
        self,
        weight_path: Union[str, List[str]] = None,
        state_dict: Union[Dict, List[Dict]] = None,
        replace_keys: Optional[Dict] = dict(),
        prefix_keys: Optional[Dict] = dict(),
    ):
        """
        Load pretrained weights into the model.

        Args:
            weight_path (str or List[str]): Path(s) to the pretrained weight file(s).
            state_dict (Dict or List[Dict]): Pretrained state_dict(s) to load weights from.
            replace_keys (Dict): Dictionary mapping keys in the pretrained state_dict to the model's keys.
            prefix_keys (Dict): Dictionary prefix keys in the pretrained state_dict to the model's keys.

        Returns:
            None
        """
        assert weight_path or state_dict, "weight_path or state_dict must be set"

        # Load state_dict(s) based on the provided weight_path or state_dict
        state_dicts = []
        if weight_path:
            if isinstance(weight_path, str):
                weight_path = [weight_path]
            for path in weight_path:
                logging.debug(f"Loading weights from {path}")
            state_dicts += [load_weight(path) for path in weight_path]

        if state_dict:
            state_dicts += state_dict if isinstance(state_dict, list) else [state_dict]

        self_state_dict = self.state_dict()  # Get the current state_dict of the model
        load_keys = []  # Keep track of the keys loaded from the state_dict(s)
        non_load_keys = []  # Keep track of the keys not loaded from the state_dict(s)

        if isinstance(self.replace_keys_in_state_dict, dict):
            replace_keys = {**self.replace_keys_in_state_dict, **replace_keys}

        if isinstance(self.prefix_keys_in_state_dict, dict):
            prefix_keys = {**self.prefix_keys_in_state_dict, **prefix_keys}

        # Iterate over the state_dict(s) and load the matching keys into the model's state_dict
        for _state_dict in state_dicts:
            if not _state_dict:
                continue
            for key, value in list(_state_dict.items()):
                for rkey, prefix in prefix_keys.items():
                    if re.match(rkey, key):
                        key = prefix + key
                        break

                for rkey, nkey in replace_keys.items():
                    key = re.sub(rkey, nkey, key)
                if key in self_state_dict and value.shape == self_state_dict[key].shape:
                    self_state_dict[key] = value
                    if key not in load_keys:
                        load_keys.append(key)
                else:
                    non_load_keys.append(key)

        self.load_state_dict(self_state_dict, False)
        load_percent = (
            len(load_keys) / len(self_state_dict) * 100
        )  # Calculate the percentage of loaded keys
        logging.debug(f"Non load keys in pretrain weights: {list(non_load_keys)}")
        logging.debug(
            f"{type(self).__name__} missed keys: {list(self_state_dict.keys() - load_keys)}"
        )
        logging.info(f"{type(self).__name__} loaded weights ({int(load_percent)}%)")


class GenericModel(nn.Module, CheckpointMixin):
    def __init__(self):
        super().__init__()
        pass

    def _init_weights(self, module):
        """
        Initialize the weights of the given module.

        Args:
            module (nn.Module): The module to initialize weights for.
        """
        if isinstance(module, (nn.Linear, nn.Embedding)):
            module.weight.data.normal_(mean=0.0, std=0.02)

        if isinstance(module, nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)

        if isinstance(module, nn.Linear) and module.bias is not None:
            module.bias.data.zero_()

    def init_weights(self):
        """
        Initialize the weights of the model.
        """
        self.apply(self._init_weights)

    @property
    def dtype(self) -> torch.dtype:
        """
        Returns the data type of the model's parameters.

        Returns:
            torch.dtype: The data type of the model's parameters.
        """
        return next(self.parameters()).dtype

    @property
    def device(self):
        """
        Returns the device of the model's parameters.

        Returns:
            torch.device: The device of the model's parameters.
        """
        return next(self.parameters()).device


from unitorch.models.processing_utils import (
    HfTextGenerationProcessor,
    HfTextClassificationProcessor,
    HfImageClassificationProcessor,
)
from unitorch.models.modeling_ema import ExponentialMovingAverage
from unitorch.models.quantization import QuantizationConfig, QuantizationMixin
from unitorch.models.onnx import GenericOnnxModel

# import models
import unitorch.models.bart
import unitorch.models.beit
import unitorch.models.bert
import unitorch.models.blip
import unitorch.models.bloom
import unitorch.models.chinese_clip
import unitorch.models.clip
import unitorch.models.deberta
import unitorch.models.dpt
import unitorch.models.llama
import unitorch.models.llava
import unitorch.models.mask2former
import unitorch.models.mbart
import unitorch.models.mistral
import unitorch.models.mt5
import unitorch.models.onnx
import unitorch.models.pegasus
import unitorch.models.roberta
import unitorch.models.swin
import unitorch.models.t5
import unitorch.models.visualbert
import unitorch.models.vit
import unitorch.models.xlm_roberta
import unitorch.models.xpegasus
import unitorch.models.peft

if is_diffusers_available():
    import unitorch.models.diffusers



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.models.mask2former.modeling import Mask2FormerForSegmentation
from unitorch.models.mask2former.processing import Mask2FormerProcessor



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import abc
from torch.multiprocessing import Process, Queue
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union

GENERATE_FINISHED = "done"
POSTPROCESS_FINISHED = None


class GenericWriter(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def process_start(self, chunk):
        pass

    @abc.abstractmethod
    def process_end(self):
        pass

    @abc.abstractmethod
    def process_chunk(self, chunk):
        pass


class IOProcess(Process):
    def __init__(self, msg_queue: Queue, writer: GenericWriter):
        super().__init__()
        self.msg_queue = msg_queue
        self.writer = writer
        self.waiting_for = 0
        self.chunk_buf = {}

    def process_start(self, chunk):
        self.writer.process_start(chunk)

    def process_end(self):
        self.writer.process_end()

    def process_chunk(self, chunk):
        self.writer.process_chunk(chunk)

    def process_buffer(self):
        while self.waiting_for in self.chunk_buf:
            self.process_chunk(self.chunk_buf[self.waiting_for])
            del self.chunk_buf[self.waiting_for]
            self.waiting_for += 1

    def run(self):
        while True:
            ind, chunk = self.msg_queue.get()
            if chunk == GENERATE_FINISHED:
                self.process_end()
                break
            if ind != self.waiting_for:
                self.chunk_buf[ind] = chunk
            else:
                if ind == 0:
                    self.process_start(chunk)
                else:
                    self.process_chunk(chunk)
                self.waiting_for += 1
                self.process_buffer()
        self.process_buffer()
        assert not self.chunk_buf, "IO Buffer not empty"
        self.msg_queue.close()
        self.msg_queue.join_thread()


class PostProcess(Process):
    def __init__(self, postprocess_fn: Callable, data_queue: Queue, msg_queue: Queue):
        super().__init__()
        self.data_queue = data_queue
        self.msg_queue = msg_queue
        self.postprocess_fn = postprocess_fn

    def run(self):
        while True:
            ind, outputs = self.data_queue.get()
            if outputs == GENERATE_FINISHED or outputs == POSTPROCESS_FINISHED:
                self.data_queue.put((-1, POSTPROCESS_FINISHED))
                break
            else:
                chunk = self.postprocess_fn(outputs)
                self.msg_queue.put((ind, chunk))

        self.data_queue.close()
        self.data_queue.join_thread()
        self.msg_queue.close()
        self.msg_queue.join_thread()



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

# pretrained infos
pretrained_dpt_infos = {
    "default-dpt": {
        "config": "https://huggingface.co/Intel/dpt-large/resolve/main/config.json",
        "vision_config": "https://huggingface.co/Intel/dpt-large/resolve/main/preprocessor_config.json",
    },
    "dpt-large": {
        "config": "https://huggingface.co/Intel/dpt-large/resolve/main/config.json",
        "vision_config": "https://huggingface.co/Intel/dpt-large/resolve/main/preprocessor_config.json",
        "weight": "https://huggingface.co/Intel/dpt-large/resolve/main/pytorch_model.bin",
    },
}

import unitorch.cli.models.dpt.modeling
import unitorch.cli.models.dpt.processing
from unitorch.cli.models.dpt.modeling import DPTForDepthEstimation
from unitorch.cli.models.dpt.processing import DPTProcessor



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import torch
import torch.nn as nn
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from unitorch.cli import (
    add_default_section_for_init,
    add_default_section_for_function,
    register_process,
)
from unitorch.cli import WriterMixin, WriterOutputs
from unitorch.cli.models.modeling_utils import ListTensorsOutputs, ListTensorsTargets


@dataclass
class DetectionOutputs(ListTensorsOutputs, WriterMixin):
    bboxes: Union[torch.Tensor, List[torch.Tensor]]
    scores: Union[torch.Tensor, List[torch.Tensor]]
    classes: Union[torch.Tensor, List[torch.Tensor]]
    features: Optional[Union[torch.Tensor, List[torch.Tensor]]] = torch.empty(0)


@dataclass
class DetectionTargets(ListTensorsTargets):
    bboxes: Union[torch.Tensor, List[torch.Tensor]]
    classes: Union[torch.Tensor, List[torch.Tensor]]
    sample_weight: Optional[torch.Tensor] = torch.empty(0)


class DetectionProcessor:
    def __init__(
        self,
    ):
        pass

    @classmethod
    @add_default_section_for_init("core/process/detection")
    def from_core_configure(cls, config, **kwargs):
        pass

    @register_process("core/postprocess/detection")
    def _detection(
        self,
        outputs: DetectionOutputs,
    ):
        results = outputs.to_pandas()
        results["bboxes"] = [b.tolist() for b in outputs.bboxes]
        results["scores"] = [s.tolist() for s in outputs.scores]
        results["classes"] = [c.tolist() for c in outputs.classes]
        return WriterOutputs(results)


def detection_model_decorator(cls):
    class DetectionModel(nn.Module):
        def __init__(self, *args, **kwargs):
            super().__init__()
            if "__detection_model__" in kwargs:
                self.model = kwargs.pop("__detection_model__")
            else:
                self.model = cls(*args, **kwargs)

            __more_attrs__ = [
                "load_state_dict",
                "state_dict",
                "save_checkpoint",
                "from_checkpoint",
                "from_pretrained",
            ]
            for __more_attr__ in __more_attrs__:
                setattr(self, __more_attr__, getattr(self.model, __more_attr__))

            assert hasattr(self.model, "detect")

        def forward(self, *args, **kwargs):
            if self.training:
                return self.model(*args, **kwargs)
            return self.model.detect(*args, **kwargs)

        @classmethod
        def from_core_configure(_cls, cfg, **kwargs):
            model = cls.from_core_configure(cfg, **kwargs)
            return _cls(__detection_model__=model)

    return DetectionModel



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import math
import logging
import torch
import torch.nn as nn
from collections import OrderedDict
from copy import deepcopy
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union


class ExponentialMovingAverage(nn.Module):
    """
    Exponential Moving Average (EMA) for model parameters.
    """

    checkpoint_name = "pytorch_ema_model.bin"

    def __init__(
        self,
        model,
        decay: Optional[float] = 0.9999,
        tau: Optional[int] = 2000,
        num_steps: Optional[int] = 0,
    ):
        """
        Initializes the ExponentialMovingAverage.

        Args:
            model (nn.Module): The model to apply EMA to.
            decay (float, optional): Decay rate for the EMA. Defaults to 0.9999.
            tau (int, optional): Time constant for the EMA. Defaults to 2000.
            num_steps (int, optional): Number of steps taken for the EMA. Defaults to 0.
        """
        super().__init__()
        self.model = deepcopy(model)
        self.num_steps = num_steps
        self.decay = lambda x: decay * (1 - math.exp(-x / tau))

        for p in self.model.parameters():
            p.requires_grad = False

    def from_checkpoint(
        self,
        ckpt_dir: str,
        weight_name: Optional[str] = None,
        **kwargs,
    ):
        """
        Load model weights from a checkpoint.

        Args:
            ckpt_dir (str): Directory path of the checkpoint.
            weight_name (str): Name of the weight file.

        Returns:
            None
        """
        if weight_name is None:
            weight_name = self.checkpoint_name
        weight_path = os.path.join(ckpt_dir, weight_name)
        if not os.path.exists(weight_path):
            return
        state_dict = torch.load(weight_path, map_location="cpu")
        self.model.load_state_dict(state_dict)
        logging.info(
            f"{type(self).__name__} model load weight from checkpoint {weight_path}"
        )

    def save_checkpoint(
        self,
        ckpt_dir: str,
        weight_name: Optional[str] = None,
        **kwargs,
    ):
        """
        Save the model's current state as a checkpoint.

        Args:
            ckpt_dir (str): Directory path to save the checkpoint.
            weight_name (str): Name of the weight file.

        Returns:
            None
        """
        if weight_name is None:
            weight_name = self.checkpoint_name
        state_dict = self.model.state_dict()
        weight_path = os.path.join(ckpt_dir, weight_name)
        torch.save(state_dict, weight_path)
        logging.info(f"{type(self).__name__} model save checkpoint to {weight_path}")

    def forward(self, *args, **kwargs):
        """
        Forward pass through the model.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            The output of the model.
        """
        return self.model(*args, **kwargs)

    @torch.no_grad()
    def step(self, model):
        """
        Performs a step of EMA.

        Args:
            model (nn.Module): The model to update the EMA with.
        """
        self.num_steps += 1
        rate = self.decay(self.num_steps)

        new_state = model.state_dict()
        for key, value in self.model.state_dict().items():
            if not value.dtype.is_floating_point:
                continue
            value *= rate
            value += (1 - rate) * new_state[key].detach()



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import torch
import torch.nn as nn
from PIL import Image
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union
from torchvision.transforms import Resize, CenterCrop, ToTensor, Normalize, Compose
from unitorch.cli.models.segmentation_utils import SegmentationTargets
from unitorch.models.bria import BRIAProcessor as _BRIAProcessor
from unitorch.cli import (
    cached_path,
    add_default_section_for_init,
    add_default_section_for_function,
    register_process,
)
from unitorch.cli.models import TensorsInputs


class BRIAProcessor(_BRIAProcessor):
    def __init__(
        self,
        image_size: Optional[int] = 1024,
    ):
        super().__init__(image_size=image_size)

    @classmethod
    @add_default_section_for_init("core/process/bria")
    def from_core_configure(cls, config, **kwargs):
        config.set_default_section("core/process/bria")
        image_size = config.getoption("image_size", 1024)

        return {
            "image_size": image_size,
        }

    @register_process("core/process/bria/segmentation/inputs")
    def _segmentation_inputs(
        self,
        image: Union[Image.Image, str],
    ):
        if isinstance(image, str):
            image = Image.open(image)
        inputs = super().segmentation_inputs(image=image)
        return TensorsInputs(images=inputs.image, sizes=inputs.sizes)

    @register_process("core/process/bria/segmentation")
    def _segmentation(
        self,
        image: Union[Image.Image, str],
        mask: Union[Image.Image, str],
    ):
        if isinstance(image, str):
            image = Image.open(image)
        if isinstance(mask, str):
            mask = Image.open(mask)
        inputs = super().segmentation_inputs(image=image)
        labels = super().segmentation_labels(image=mask)
        return TensorsInputs(images=inputs.image), SegmentationTargets(
            targets=labels.image
        )
